import uuid

from core.models import MutationLog
from invoice.models import (
    Invoice,
    InvoiceLineItem,
    PaymentInvoice,
    PaymentInvoiceMutation
)
from invoice.tests.gql.base import InvoiceGQLTestCase
from invoice.tests.helpers import (
    DEFAULT_TEST_PAYMENT_INVOICE_PAYLOAD,
    create_test_invoice,
    create_test_invoice_line_item,
    create_test_payment_invoice_with_details,
)


class PaymentInvoiceGQLTest(InvoiceGQLTestCase):

    search_for_payment_invoice_query = F'''
query {{ 
	paymentInvoice(codeTp_Iexact:"{DEFAULT_TEST_PAYMENT_INVOICE_PAYLOAD['code_tp']}", 
	amountReceived: "{DEFAULT_TEST_PAYMENT_INVOICE_PAYLOAD['amount_received']}"){{
    edges {{
      node {{
        isDeleted,
        codeTp,
        codeExt,
        amountReceived,
        datePayment,
        reconciliationStatus,
        fees,
        payerRef
      }}
    }}
  }}
}}
'''

    create_mutation_str = '''  
mutation {{
  createPaymentInvoice(input:{{reconciliationStatus: 1, codeExt:"{payment_code}", codeTp:"PAY_CODE", codeReceipt:"gqlRec", 
  label:"gql label", fees: "12.00", amountReceived: "91.50", payerRef: "payerRef", 
  datePayment:"2022-04-12", clientMutationId: "{mutation_id}"}}) {{
    internalId
    clientMutationId
  }}
}}
'''

    delete_mutation_str = '''  
mutation {{
  deletePaymentInvoice(input:{{uuids:["{payment_uuid}"], clientMutationId: "{mutation_id}"}}) {{
    internalId
    clientMutationId
  }}
}}
'''

    update_mutation_str = '''  
mutation {{
	updatePaymentInvoice(input:{{id:"{payment_uuid}", codeExt:"updExt", payerRef: "payerRef", clientMutationId: "{mutation_id}"}}){{
    internalId
    clientMutationId
  }}
}}
'''

    search_for_payment_invoice_with_detail_query = F'''
query {{ 
    paymentInvoice(codeTp_Iexact:"{DEFAULT_TEST_PAYMENT_INVOICE_PAYLOAD['code_tp']}", 
    amountReceived: "{DEFAULT_TEST_PAYMENT_INVOICE_PAYLOAD['amount_received']}"){{
    edges {{
        node {{
        isDeleted,
        codeTp,
        codeExt,
        amountReceived,
        datePayment,
        reconciliationStatus,
        fees,
        payerRef,
        invoicePayments{{
          totalCount
          edges{{
            node{{
              subjectTypeName
              fees
              amount
              status
            }}
          }}
        }}
        }}
    }}
    }}
}}
'''

    create_mutation_with_detail_str = '''  
mutation {{
    createPaymentWithDetailInvoice(input:{{status: 1, subjectId: "{invoice_uuid}", subjectType: "invoice"
    reconciliationStatus: 1, codeExt:"{payment_code}", codeTp:"PAY_CODE", codeReceipt:"gqlRec", 
    label:"gql label", fees: "12.00", amountReceived: "91.50", payerRef: "payerRef", 
    datePayment:"2022-04-12", clientMutationId: "{mutation_id}"}}) {{
        internalId
        clientMutationId
    }}
}}
'''

    def test_fetch_payment_invoice_query(self):
        payment_code = "GQLCOD"
        mutation_client_id = str(uuid.uuid4())
        mutation = self.create_mutation_str.format(
            payment_code=payment_code, mutation_id=mutation_client_id
        )
        self.graph_client.execute(mutation, context=self.BaseTestContext(self.user))
        output = self.graph_client.execute(self.search_for_payment_invoice_query,
                                           context=self.BaseTestContext(self.user))
        expected = \
            {'data': {
                'paymentInvoice': {
                    'edges': [
                        {'node': {
                            'isDeleted': False,
                            'codeExt': 'GQLCOD',
                            'codeTp': 'PAY_CODE',
                            'amountReceived': '91.50',
                            'datePayment': '2022-04-12',
                            'reconciliationStatus': 'A_1',
                            'fees': '12.00',
                            'payerRef': 'payerRef',
        }}]}}}
        self.assertEqual(output, expected)

    def test_fetch_payment_invoice_with_detail_query(self):
        payment = create_test_payment_invoice_with_details()
        output = self.graph_client.execute(self.search_for_payment_invoice_with_detail_query,
                                           context=self.BaseTestContext(self.user))
        expected = \
            {'data': {
                'paymentInvoice': {
                    'edges': [
                        {'node': {
                            'isDeleted': False,
                            'codeExt': payment.code_ext,
                            'codeTp': payment.code_tp,
                            'amountReceived': '91.50',
                            'datePayment': '2022-04-11',
                            'reconciliationStatus': 'A_0',
                            'fees': '12.00',
                            'payerRef': payment.payer_ref,
                            'invoicePayments': {
                                'totalCount': 1,
                                'edges': [
                                    {
                                        'node': {
                                            'subjectTypeName': 'invoice',
                                            'fees': '12.00',
                                            'amount': '91.50',
                                            'status': 'A_1'
                                        }
                                    }
                                ]
                            }
        }}]}}}
        self.assertEqual(output, expected)

    def test_create_payment_mutation(self):
        payment_code = "GQLCOD"
        mutation_client_id = str(uuid.uuid4())
        mutation = self.create_mutation_str.format(
            payment_code=payment_code, mutation_id=mutation_client_id
        )
        self.graph_client.execute(mutation, context=self.BaseTestContext(self.user))
        expected = PaymentInvoice.objects.get(code_ext=payment_code)
        mutation_log = MutationLog.objects.filter(client_mutation_id=mutation_client_id).first()
        obj = PaymentInvoiceMutation.objects.get(mutation_id=mutation_log.id).payment_invoice
        self.assertEqual(obj, expected)

    def test_create_payment_with_detail_mutation(self):
        invoice = create_test_invoice()
        invoice_item = create_test_invoice_line_item(invoice)
        payment_code = "GQLCOD"
        mutation_client_id = str(uuid.uuid4())
        mutation = self.create_mutation_with_detail_str.format(
            payment_code=payment_code, mutation_id=mutation_client_id, invoice_uuid=invoice.id
        )
        self.graph_client.execute(mutation, context=self.BaseTestContext(self.user))
        expected = PaymentInvoice.objects.get(code_ext=payment_code)
        mutation_log = MutationLog.objects.filter(client_mutation_id=mutation_client_id).first()
        obj = PaymentInvoiceMutation.objects.get(mutation_id=mutation_log.id).payment_invoice
        payment_invoice = PaymentInvoice.objects.filter(invoice_payments__subject_id__in=[invoice.id]).first()
        self.assertEqual(obj, expected)
        self.assertEqual(obj, payment_invoice)
        InvoiceLineItem.objects.filter(id=invoice_item.id).delete()
        Invoice.objects.filter(id=invoice.id).delete()

    def test_delete_payment_mutation(self):
        payment_code = "GQLCOD"
        mutation_client_id = str(uuid.uuid4())
        mutation = self.create_mutation_str.format(
            payment_code=payment_code, mutation_id=mutation_client_id
        )
        self.graph_client.execute(mutation, context=self.BaseTestContext(self.user))
        expected = PaymentInvoice.objects.get(code_ext=payment_code)
        mutation_client_id = str(uuid.uuid4())
        mutation = self.delete_mutation_str.format(payment_uuid=expected.id, mutation_id=mutation_client_id)
        self.graph_client.execute(mutation, context=self.BaseTestContext(self.user))
        # TODO: Currently deleted entries are not filtered by manager, only in GQL Query. Should we change this?
        payment = PaymentInvoice.objects.filter(code_ext=payment_code).all()
        mutation_ = PaymentInvoiceMutation.objects.filter(payment_invoice=payment[0]).all()
        self.assertEqual(len(payment), 1)
        self.assertTrue(payment[0].is_deleted)
        self.assertTrue(len(mutation_) == 2)

    def test_update_payment_mutation(self):
        payment_code = "GQLCOD"
        mutation_client_id = str(uuid.uuid4())
        mutation = self.create_mutation_str.format(
            payment_code=payment_code, mutation_id=mutation_client_id
        )
        self.graph_client.execute(mutation, context=self.BaseTestContext(self.user))

        created = PaymentInvoice.objects.get(code_ext=payment_code)
        mutation_client_id = str(uuid.uuid4())
        mutation = self.update_mutation_str.format(payment_uuid=created.id, mutation_id=mutation_client_id)
        self.graph_client.execute(mutation, context=self.BaseTestContext(self.user))

        expected_code_ext = "updExt"
        mutation_log = MutationLog.objects.filter(client_mutation_id=mutation_client_id).first()
        obj: PaymentInvoice = PaymentInvoiceMutation.objects.get(mutation_id=mutation_log.id).payment_invoice
        self.assertEqual(obj.code_ext, expected_code_ext)



from datetime import date

from invoice.models import Invoice, InvoicePayment, Bill, BillPayment, PaymentInvoice, DetailPaymentInvoice

DEFAULT_TEST_INVOICE_LINE_ITEM_PAYLOAD = {
    'code': 'LineItem1',
    'description': 'description_str',
    'details': {"test_int": 1, "test_txt": "some_str"},
    'ledger_account': 'account',
    'quantity': 10,
    'unit_price': 10.5,
    'discount': 15.5,
    'tax_rate': None,
    'tax_analysis': {'lines': [{'code': 'c', 'label': 'l', 'base': '0.1', 'amount': '2.00'}], 'total': '2.0'},
    'amount_net': 10 * 10.5 - 15.5,
    'amount_total': (10 * 10.5 - 15.5) + 2.0
}

DEFAULT_TEST_INVOICE_PAYLOAD = {
    'subject_type': 'contract',
    'subject_id': None,
    'thirdparty_type': 'insuree',
    'thirdparty_id': None,
    'code': 'INVOICE_CODE',
    'code_tp': 'INVOICE_CODE_TP',
    'code_ext': 'INVOICE_CODE_EXT',
    'date_due': date(2021, 9, 13),
    'date_invoice': date(2021, 9, 11),
    'date_payed': date(2021, 9, 12),
    'amount_discount': 20.1,
    'amount_net': 20.1,
    'tax_analysis': {'lines': [{'code': 'c', 'label': 'l', 'base': '0.1', 'amount': '2.01'}], 'total': '2.01'},
    'amount_total': 20.1,
    'status': Invoice.Status.VALIDATED,  # Validated
    'note': 'NOTE',
    'terms': 'TERMS',
    'payment_reference': 'payment reference'
}

DEFAULT_TEST_INVOICE_PAYMENT_PAYLOAD = {
    'label': 'label_pay',
    'code_tp': 'pay_sys_ref',
    'code_receipt': 'receipt number',
    'invoice': None,
    'amount_payed': '91.50',
    'fees': '12.00',
    'amount_received': '22.00',
    'date_payment': date(2021, 10, 10),
    'status': InvoicePayment.PaymentStatus.ACCEPTED
}

DEFAULT_TEST_BILL_PAYLOAD = {
    'subject_type': 'contract',
    'subject_id': None,
    'thirdparty_type': 'insuree',
    'thirdparty_id': None,
    'code': 'BILL_CODE',
    'code_tp': 'BILL_CODE_TP',
    'code_ext': 'BILL_CODE_EXT',
    'date_due': date(2021, 9, 13),
    'date_bill': date(2021, 9, 11),
    'date_payed': date(2021, 9, 12),
    'amount_discount': 20.1,
    'amount_net': 20.1,
    'tax_analysis': {'lines': [{'code': 'c', 'label': 'l', 'base': '0.1', 'amount': '2.01'}], 'total': '2.01'},
    'amount_total': 20.1,
    'status': Bill.Status.VALIDATED,  # Validated
    'note': 'NOTE',
    'terms': 'TERMS',
    'payment_reference': 'payment reference'
}

DEFAULT_TEST_BILL_PAYMENT_PAYLOAD = {
    'label': 'label_pay',
    'code_tp': 'pay_sys_ref',
    'code_receipt': 'receipt number',
    'bill': None,
    'amount_payed': 91.5,
    'fees': 12.0,
    'amount_received': 22.0,
    'date_payment': date(2021, 10, 10),
    'status': BillPayment.PaymentStatus.ACCEPTED
}

DEFAULT_TEST_BILL_LINE_ITEM_PAYLOAD = {
    'code': 'LineItem1',
    'description': 'description_str',
    'details': {"test_int": 1, "test_txt": "some_str"},
    'ledger_account': 'account',
    'quantity': 10,
    'unit_price': 10.5,
    'discount': 15.5,
    'tax_rate': None,
    'tax_analysis': {'lines': [{'code': 'c', 'label': 'l', 'base': '0.1', 'amount': '2.00'}], 'total': '2.0'},
    'amount_net': 10 * 10.5 - 15.5,
    'amount_total': (10 * 10.5 - 15.5) + 2.0
}

DEFAULT_TEST_PAYMENT_INVOICE_PAYLOAD = {
    'code_tp': 'PAY_CODE',
    'code_ext': 'PAY_CODE_EXT',
    'code_receipt': 'PAY_CODE_RCP',
    'label': 'test label',
    'reconciliation_status': PaymentInvoice.ReconciliationStatus.NOT_RECONCILIATED,
    'fees': 12.00,
    'amount_received': 91.50,
    'date_payment': date(2022, 4, 11),
    'payment_origin': 'payment origin',
    'payer_ref': 'payment reference',
    'payer_name': 'payer name'
}

DEFAULT_TEST_DETAIL_PAYMENT_INVOICE_PAYLOAD = {
    'payment_id': None,
    'subject_type': 'invoice',
    'subject_id': None,
    'status': DetailPaymentInvoice.DetailPaymentStatus.ACCEPTED,
    'fees': 12.00,
    'amount': 91.50,
    'reconcilation_id': 'RECONCID',
    'reconcilation_date': date(2022, 4, 11),
}



from graphene import Schema
from graphene.test import Client
from policy.test_helpers import create_test_policy
from product.test_helpers import create_test_product

from core.forms import User
from django.test import TestCase

from contract.tests.helpers import create_test_contract
from policyholder.tests.helpers import create_test_policy_holder
from insuree.test_helpers import create_test_insuree

from invoice import schema as invoice_schema
from invoice.tests.helpers import create_test_invoice, \
    create_test_invoice_line_item
from invoice.tests.helpers.invoice_payment_helpers import create_test_invoice_payment


class InvoiceGQLTestCase(TestCase):
    class BaseTestContext:
        def __init__(self, user):
            self.user = user

    @classmethod
    def setUpClass(cls):
        super(InvoiceGQLTestCase, cls).setUpClass()
        cls._graphene_setup()

        cls.maxDiff = None
        if not User.objects.filter(username='admin_invoice').exists():
            User.objects.create_superuser(username='admin_invoice', password='S\/pe®Pąßw0rd™')

        cls.user = User.objects.filter(username='admin_invoice').first()

        cls.policy_holder = create_test_policy_holder()
        cls.contract = create_test_contract(cls.policy_holder)
        cls.insuree = create_test_insuree(with_family=True)
        cls.product = create_test_product("TestC0d3", custom_props={"insurance_period": 12})
        cls.policy = create_test_policy(
            product=cls.product,
            insuree=cls.insuree
        )

        cls.invoice = create_test_invoice(cls.contract, cls.insuree)
        cls.invoice_line_item = \
            create_test_invoice_line_item(invoice=cls.invoice, line_item=cls.policy, user=cls.user)
        cls.invoice_payment = create_test_invoice_payment(invoice=cls.invoice, user=cls.user)

    @classmethod
    def _graphene_setup(cls):
        cls.schema = Schema(
            query=invoice_schema.Query,
            mutation=invoice_schema.Mutation
        )
        cls.graph_client = Client(cls.schema)



# Generated by Django 3.0.14 on 2022-04-11 10:23

import core.fields
import dirtyfields.dirtyfields
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import invoice.mixins
import simple_history.models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0002_remove_content_type_name'),
        ('invoice', '0005_invoice_bill_roles_for_accountant'),
    ]

    operations = [
        migrations.CreateModel(
            name='PaymentInvoice',
            fields=[
                ('id', models.UUIDField(db_column='UUID', default=None, editable=False, primary_key=True, serialize=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', models.JSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', core.fields.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', core.fields.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('code_tp', models.CharField(db_column='CodeTp', max_length=255, null=True)),
                ('code_ext', models.CharField(db_column='CodeExt', max_length=255, null=True)),
                ('code_receipt', models.CharField(db_column='CodeReceipt', max_length=255, null=True)),
                ('label', models.CharField(db_column='Label', max_length=255, null=True)),
                ('reconciliation_status', models.SmallIntegerField(choices=[(0, 'rejected'), (1, 'accepted'), (2, 'refunded'), (3, 'cancelled')], db_column='ReconciliationStatus')),
                ('fees', models.DecimalField(db_column='Fees', decimal_places=2, max_digits=18, null=True)),
                ('amount_received', models.DecimalField(db_column='AmountReceived', decimal_places=2, max_digits=18, null=True)),
                ('date_payment', core.fields.DateField(db_column='DatePayment', null=True)),
                ('payment_origin', models.CharField(db_column='PaymentOrigin', max_length=255, null=True)),
                ('payer_ref', models.CharField(db_column='PayerRef', max_length=255)),
                ('payer_name', models.CharField(db_column='PayerName', max_length=255, null=True)),
                ('user_created', models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='paymentinvoice_user_created', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='paymentinvoice_user_updated', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'tblPaymentInvoice',
                'managed': True,
            },
            bases=(invoice.mixins.GenericInvoiceQuerysetMixin, dirtyfields.dirtyfields.DirtyFieldsMixin, models.Model),
        ),
        migrations.CreateModel(
            name='HistoricalPaymentInvoice',
            fields=[
                ('id', models.UUIDField(db_column='UUID', db_index=True, default=None, editable=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', models.JSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', core.fields.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', core.fields.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('code_tp', models.CharField(db_column='CodeTp', max_length=255, null=True)),
                ('code_ext', models.CharField(db_column='CodeExt', max_length=255, null=True)),
                ('code_receipt', models.CharField(db_column='CodeReceipt', max_length=255, null=True)),
                ('label', models.CharField(db_column='Label', max_length=255, null=True)),
                ('reconciliation_status', models.SmallIntegerField(choices=[(0, 'rejected'), (1, 'accepted'), (2, 'refunded'), (3, 'cancelled')], db_column='ReconciliationStatus')),
                ('fees', models.DecimalField(db_column='Fees', decimal_places=2, max_digits=18, null=True)),
                ('amount_received', models.DecimalField(db_column='AmountReceived', decimal_places=2, max_digits=18, null=True)),
                ('date_payment', core.fields.DateField(db_column='DatePayment', null=True)),
                ('payment_origin', models.CharField(db_column='PaymentOrigin', max_length=255, null=True)),
                ('payer_ref', models.CharField(db_column='PayerRef', max_length=255)),
                ('payer_name', models.CharField(db_column='PayerName', max_length=255, null=True)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('user_created', models.ForeignKey(blank=True, db_column='UserCreatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(blank=True, db_column='UserUpdatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'historical payment invoice',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='HistoricalDetailPaymentInvoice',
            fields=[
                ('id', models.UUIDField(db_column='UUID', db_index=True, default=None, editable=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', models.JSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', core.fields.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', core.fields.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('subject_id', models.CharField(db_column='SubjectId', max_length=255, null=True)),
                ('status', models.SmallIntegerField(choices=[(0, 'rejected'), (1, 'accepted'), (2, 'refunded'), (3, 'cancelled')], db_column='Status')),
                ('fees', models.DecimalField(db_column='Fees', decimal_places=2, max_digits=18, null=True)),
                ('amount', models.DecimalField(db_column='Amount', decimal_places=2, max_digits=18, null=True)),
                ('reconcilation_id', models.CharField(db_column='ReconcilationId', max_length=255, null=True)),
                ('reconcilation_date', models.DateField(db_column='ReconcilationDate', null=True)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('payment', models.ForeignKey(blank=True, db_column='PaymentUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='invoice.PaymentInvoice')),
                ('subject_type', models.ForeignKey(blank=True, db_column='SubjectType', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='contenttypes.ContentType')),
                ('user_created', models.ForeignKey(blank=True, db_column='UserCreatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(blank=True, db_column='UserUpdatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'historical detail payment invoice',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='DetailPaymentInvoice',
            fields=[
                ('id', models.UUIDField(db_column='UUID', default=None, editable=False, primary_key=True, serialize=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', models.JSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', core.fields.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', core.fields.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('subject_id', models.CharField(db_column='SubjectId', max_length=255, null=True)),
                ('status', models.SmallIntegerField(choices=[(0, 'rejected'), (1, 'accepted'), (2, 'refunded'), (3, 'cancelled')], db_column='Status')),
                ('fees', models.DecimalField(db_column='Fees', decimal_places=2, max_digits=18, null=True)),
                ('amount', models.DecimalField(db_column='Amount', decimal_places=2, max_digits=18, null=True)),
                ('reconcilation_id', models.CharField(db_column='ReconcilationId', max_length=255, null=True)),
                ('reconcilation_date', models.DateField(db_column='ReconcilationDate', null=True)),
                ('payment', models.ForeignKey(db_column='PaymentUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='payment_invoice', to='invoice.PaymentInvoice')),
                ('subject_type', models.ForeignKey(db_column='SubjectType', null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='contenttypes.ContentType')),
                ('user_created', models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='detailpaymentinvoice_user_created', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='detailpaymentinvoice_user_updated', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'tblDetailPaymentInvoice',
                'managed': True,
            },
            bases=(invoice.mixins.GenericInvoiceQuerysetMixin, dirtyfields.dirtyfields.DirtyFieldsMixin, models.Model),
        ),
    ]



import graphene
from django.contrib.auth.models import AnonymousUser
from django.db.models import Q

from core.schema import OrderedDjangoFilterConnectionField
from core.utils import append_validity_filter
from invoice.apps import InvoiceConfig
from invoice.gql.gql_types.bill_types import BillItemGQLType
from invoice.models import BillItem, Bill
import graphene_django_optimizer as gql_optimizer


class BillItemQueryMixin:
    bill_item = OrderedDjangoFilterConnectionField(
        BillItemGQLType,
        orderBy=graphene.List(of_type=graphene.String),
        dateValidFrom__Gte=graphene.DateTime(),
        dateValidTo__Lte=graphene.DateTime(),
        applyDefaultValidityFilter=graphene.Boolean(),
        client_mutation_id=graphene.String(),
    )

    def resolve_bill_item(self, info, **kwargs):
        BillItemQueryMixin._check_permissions(info.context.user)
        filters = []
        filters += append_validity_filter(**kwargs)

        client_mutation_id = kwargs.get("client_mutation_id", None)
        if client_mutation_id:
            filters.append(Q(mutations__mutation__client_mutation_id=client_mutation_id))

        line_type = kwargs.get("line_type", None)
        if line_type:
            filters.append(Q(line_type__model=line_type))

        bill_li_qs = BillItem.objects.filter(*filters)

        if InvoiceConfig.bill_user_filter:
            bill_qs = InvoiceConfig.bill_user_filter(Bill.objects.all(), info.context.user)
            bill_li_qs = bill_li_qs.filter(bill__in=bill_qs)

        return gql_optimizer.query(bill_li_qs, info)

    @staticmethod
    def _check_permissions(user):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(
                InvoiceConfig.gql_bill_search_perms):
            raise PermissionError("Unauthorized")



import graphene
from django.utils.translation import gettext as _
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ValidationError
from django.db.models import Q

from core.gql.gql_mutations.base_mutation import BaseHistoryModelCreateMutationMixin, BaseMutation, \
    BaseHistoryModelUpdateMutationMixin, BaseHistoryModelDeleteMutationMixin
from core.schema import OpenIMISMutation
from invoice.apps import InvoiceConfig
from invoice.gql.input_types import CreateInvoiceEventType, UpdateInvoiceEventType
from invoice.models import InvoiceEvent, InvoiceEventMutation


class CreateInvoiceEventMutation(BaseHistoryModelCreateMutationMixin, BaseMutation):
    _mutation_class = "CreateInvoiceEventMutation"
    _mutation_module = "invoice"
    _model = InvoiceEvent

    @classmethod
    def _mutate(cls, user, **data):
        client_mutation_id = data.get("client_mutation_id")
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')
        p = cls.create_object(user=user, object_data=data)
        if p:
            InvoiceEventMutation.object_mutated(
                user,
                client_mutation_id=client_mutation_id,
                invoice_event=p
            )

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(
                InvoiceConfig.gql_invoice_event_create_message_perms):
            raise PermissionError("Unauthorized")

    class Input(CreateInvoiceEventType):
        pass


class DeleteUserInvoiceEventMutation(BaseHistoryModelDeleteMutationMixin, BaseMutation):
    _mutation_class = "DeleteUserInvoiceEventMutation"
    _mutation_module = "invoice"
    _model = InvoiceEvent

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(
                InvoiceConfig.gql_invoice_event_delete_my_message_perms):
            raise PermissionError("Unauthorized")

        objs_ = cls._model.objects.filter(id=data['uuids'])
        not_created_by_user = objs_.filter(~Q(user_created=user))
        if not_created_by_user.exists():
            raise PermissionError(
                _(F"User is not creator of invoice events: %(events)s, events can't be deleted.")
                % {'events': str(not_created_by_user.values_list('uuid', flat=True))}
            )

    class Input(OpenIMISMutation.Input):
        uuids = graphene.List(graphene.UUID)




from django.conf import settings
from django.db import models
from graphql import ResolveInfo

from core.models import HistoryModelManager


class GenericInvoiceManager(HistoryModelManager):
    def filter(self, *args, **kwargs):
        keys = [x for x in kwargs if "itemsvc" in x]
        for key in keys:
            new_key = key.replace("itemsvc", self.model.model_prefix)
            kwargs[new_key] = kwargs.pop(key)
        return super(GenericInvoiceManager, self).filter(*args, **kwargs)


class GenericInvoiceQuerysetMixin:

    @classmethod
    def get_queryset(cls, queryset, user):
        queryset = cls.filter_queryset(queryset)
        if isinstance(user, ResolveInfo):
            user = user.context.user
        if settings.ROW_SECURITY and user.is_anonymous:
            return queryset.filter(id=-1)
        if settings.ROW_SECURITY:
            pass
        return queryset



from django.db import transaction

from core.services import BaseService
from core.services.utils import (
    check_authentication,
    output_exception,
    output_result_success,
    model_representation
)
from core.signals import *
from invoice.models import (
    Bill,
    PaymentInvoice,
    Invoice,
    DetailPaymentInvoice
)
from invoice.utils import resolve_payment_details
from invoice.validation.paymentInvoice import PaymentInvoiceModelValidation


class PaymentInvoiceService(BaseService):

    OBJECT_TYPE = PaymentInvoice

    def __init__(self, user, validation_class: PaymentInvoiceModelValidation = PaymentInvoiceModelValidation):
        super().__init__(user, validation_class)
        self.validation_class = validation_class

    @check_authentication
    def update(self, obj_data):
        raise NotImplementedError("Update method is not implemented for PaymentInvoice")

    @check_authentication
    def create_with_detail(self, payment_invoice: dict, payment_detail: DetailPaymentInvoice):
        try:
            with transaction.atomic():
                payment = PaymentInvoice(**payment_invoice)
                payment.save(username=self.user.username)
                payment_detail.payment = payment
                payment_detail.subject = self._get_generic_object(
                    payment_detail.subject_id,
                    payment_detail.subject_type
                )
                if 'reconciliation' in payment.json_ext:
                    payment_detail.reconcilation_id = payment.json_ext['reconciliation']['id']
                payment_detail.save(username=self.user.username)
                dict_repr = model_representation(payment)
                dict_repr['payment_detail_uuid'] = payment_detail.uuid
                return output_result_success(dict_representation=dict_repr)
        except Exception as exc:
            return output_exception(model_name="PaymentInvoice", method="create_with_detail", exception=exc)

    @check_authentication
    @register_service_signal('signal_after_invoice_module_ref_received')
    def ref_received(self, payment_invoice: PaymentInvoice, payment_ref):
        try:
            with transaction.atomic():
                self.validation_class.validate_ref_received(self.user, payment_invoice, payment_ref)
                payment_invoice.code_ext = payment_ref
                return self.save_instance(payment_invoice)
        except Exception as exc:
            return output_exception(model_name="PaymentInvoice", method="ref_received", exception=exc)

    @register_service_signal('signal_after_invoice_module_payment_received')
    def payment_received(self, payment_invoice: PaymentInvoice, payment_status: DetailPaymentInvoice.DetailPaymentStatus):
        try:
            with transaction.atomic():
                self.validation_class.validate_receive_payment(self.user, payment_invoice)
                self._update_all_dependencies_for_payment(
                    payment_invoice,
                    payment_status,
                    Invoice.Status.PAID
                )
                dict_repr = model_representation(payment_invoice)
                return output_result_success(dict_representation=dict_repr)
        except Exception as exc:
            return output_exception(model_name="PaymentInvoice", method="payment_received", exception=exc)

    @register_service_signal('signal_after_invoice_module_payment_refunded')
    def payment_refunded(self, payment_invoice):
        try:
            with transaction.atomic():
                self._update_all_dependencies_for_payment(
                    payment_invoice,
                    DetailPaymentInvoice.DetailPaymentStatus.REFUNDED,
                    Invoice.Status.SUSPENDED
                )
                dict_repr = model_representation(payment_invoice)
                return output_result_success(dict_representation=dict_repr)
        except Exception as exc:
            return output_exception(model_name="PaymentInvoice", method="payment_refunded", exception=exc)

    @register_service_signal('signal_after_invoice_module_payment_cancelled')
    def payment_cancelled(self, payment_invoice):
        try:
            with transaction.atomic():
                self._update_all_dependencies_for_payment(
                    payment_invoice,
                    DetailPaymentInvoice.DetailPaymentStatus.CANCELLED,
                    Invoice.Status.SUSPENDED
                )
                dict_repr = model_representation(payment_invoice)
                return output_result_success(dict_representation=dict_repr)
        except Exception as exc:
            return output_exception(model_name="PaymentInvoice", method="payment_cancelled", exception=exc)

    def _update_all_dependencies_for_payment(self, payment_invoice, payment_status, invoice_status):
        invoices, bills = resolve_payment_details(payment_invoice)
        payment_details = payment_invoice.invoice_payments.all()
        self._update_detail(payment_details, payment_status)
        self._update_detail(invoices, invoice_status)
        self._update_detail(bills, invoice_status)

    def _update_detail(self, detail_collection, status):
        for detail in detail_collection:
            if detail.status != status:
                detail.status = status
                detail.save(username=self.user.username)

    @classmethod
    def _get_generic_object(cls, subject_id, subject_type):
        if subject_type.model == 'invoice':
            object = Invoice.objects.get(id=subject_id)
        else:
            object = Bill.objects.get(id=subject_id)
        return object



import uuid

from core.models import MutationLog
from invoice.models import (
    DetailPaymentInvoice,
    DetailPaymentInvoiceMutation,
    Invoice,
    InvoiceLineItem
)
from invoice.tests.gql.base import InvoiceGQLTestCase
from invoice.tests.helpers import (
    DEFAULT_TEST_PAYMENT_INVOICE_PAYLOAD,
    create_test_payment_invoice_without_details
)


class DetailPaymentInvoiceGQLTest(InvoiceGQLTestCase):
    search_for_detail_payment_invoice_query = F'''
query {{ 
	detailPaymentInvoice(payment_CodeTp_Iexact:"{DEFAULT_TEST_PAYMENT_INVOICE_PAYLOAD['code_tp']}"){{
    edges {{
      node {{
        isDeleted,
        amount,
        status,
        fees,
        subjectTypeName,
        subjectId
      }}
    }}
  }}
}}
'''

    create_mutation_str = '''  
mutation {{
  createDetailPaymentInvoice(input:{{status: 1, fees: "12.00", amount: "91.50", 
  paymentId: "{payment_uuid}", subjectType: "invoice", subjectId: "{subject_uuid}", 
  clientMutationId: "{mutation_id}"}}) {{
    internalId
    clientMutationId
  }}
}}
'''

    delete_mutation_str = '''  
mutation {{
  deleteDetailPaymentInvoice(input:{{uuids:["{payment_uuid}"], clientMutationId: "{mutation_id}"}}) {{
    internalId
    clientMutationId
  }}
}}
'''

    update_mutation_str = '''  
mutation {{
	updateDetailPaymentInvoice(input:{{id:"{detail_payment_uuid}", 
	  status: 1, fees: "12.00", amount: "91.50", 
      clientMutationId: "{mutation_id}"}}){{
    internalId
    clientMutationId
  }}
}}
'''

    def test_fetch_detail_payment_invoice_query(self):
        payment, invoice, invoice_item = create_test_payment_invoice_without_details()
        mutation_client_id = str(uuid.uuid4())
        mutation = self.create_mutation_str.format(
            payment_uuid=payment.id, subject_uuid=invoice.id, mutation_id=mutation_client_id
        )
        self.graph_client.execute(mutation, context=self.BaseTestContext(self.user))
        output = self.graph_client.execute(self.search_for_detail_payment_invoice_query,
                                           context=self.BaseTestContext(self.user))
        expected = \
            {'data': {
                'detailPaymentInvoice': {
                    'edges': [
                        {'node': {
                            'amount': '91.50',
                            'fees': '12.00',
                            'isDeleted': False,
                            'status': 'A_1',
                            'subjectId': f"{invoice.id}",
                            'subjectTypeName': "invoice",
                        }}]}}}
        self.assertEqual(output, expected)
        InvoiceLineItem.objects.filter(id=invoice_item.id).delete()
        Invoice.objects.filter(id=invoice.id).delete()

    def test_create_detail_payment_mutation(self):
        payment, invoice, invoice_item = create_test_payment_invoice_without_details()
        mutation_client_id = str(uuid.uuid4())
        mutation = self.create_mutation_str.format(
            payment_uuid=payment.id, subject_uuid=invoice.id, mutation_id=mutation_client_id
        )
        self.graph_client.execute(mutation, context=self.BaseTestContext(self.user))
        expected = DetailPaymentInvoice.objects.get(payment__id=payment.id)
        mutation_log = MutationLog.objects.filter(client_mutation_id=mutation_client_id).first()
        obj = DetailPaymentInvoiceMutation.objects.get(mutation_id=mutation_log.id).detail_payment_invoice
        self.assertEqual(obj, expected)
        InvoiceLineItem.objects.filter(id=invoice_item.id).delete()
        Invoice.objects.filter(id=invoice.id).delete()

    def test_delete_detail_payment_mutation(self):
        payment, invoice, invoice_item = create_test_payment_invoice_without_details()
        mutation_client_id = str(uuid.uuid4())
        mutation = self.create_mutation_str.format(
            payment_uuid=payment.id, subject_uuid=invoice.id, mutation_id=mutation_client_id
        )
        self.graph_client.execute(mutation, context=self.BaseTestContext(self.user))
        expected = DetailPaymentInvoice.objects.get(payment__id=payment.id)
        mutation_client_id = str(uuid.uuid4())
        mutation = self.delete_mutation_str.format(payment_uuid=expected.id, mutation_id=mutation_client_id)
        self.graph_client.execute(mutation, context=self.BaseTestContext(self.user))
        # TODO: Currently deleted entries are not filtered by manager, only in GQL Query. Should we change this?
        detail_payment = DetailPaymentInvoice.objects.filter(payment__id=payment.id).all()
        mutation_ = DetailPaymentInvoiceMutation.objects.filter(detail_payment_invoice=detail_payment[0]).all()
        self.assertEqual(len(detail_payment), 1)
        self.assertTrue(detail_payment[0].is_deleted)
        self.assertTrue(len(mutation_) == 2)
        InvoiceLineItem.objects.filter(id=invoice_item.id).delete()
        Invoice.objects.filter(id=invoice.id).delete()

    def test_update_detail_payment_mutation(self):
        payment, invoice, invoice_item = create_test_payment_invoice_without_details()
        mutation_client_id = str(uuid.uuid4())
        mutation = self.create_mutation_str.format(
            payment_uuid=payment.id, subject_uuid=invoice.id, mutation_id=mutation_client_id
        )
        self.graph_client.execute(mutation, context=self.BaseTestContext(self.user))

        created = DetailPaymentInvoice.objects.get(payment__id=payment.id)
        mutation_client_id = str(uuid.uuid4())
        mutation = self.update_mutation_str.format(detail_payment_uuid=created.id, mutation_id=mutation_client_id)
        self.graph_client.execute(mutation, context=self.BaseTestContext(self.user))

        expected_fees = "12.00"
        mutation_log = MutationLog.objects.filter(client_mutation_id=mutation_client_id).first()
        obj: DetailPaymentInvoice = DetailPaymentInvoiceMutation.objects.get(
            mutation_id=mutation_log.id).detail_payment_invoice
        self.assertEqual(str(obj.fees), str(expected_fees))
        InvoiceLineItem.objects.filter(id=invoice_item.id).delete()
        Invoice.objects.filter(id=invoice.id).delete()



# Generated by Django 3.0.14 on 2021-10-27 08:30

import core.fields
import core.models
import datetime
import dirtyfields.dirtyfields
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import invoice.mixins
import invoice.models
import simple_history.models
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('core', '0015_missing_roles'),
    ]

    operations = [
        migrations.CreateModel(
            name='Bill',
            fields=[
                ('id', models.UUIDField(db_column='UUID', default=None, editable=False, primary_key=True, serialize=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', models.JSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', core.fields.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', core.fields.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('date_valid_from', core.fields.DateTimeField(db_column='DateValidFrom', default=datetime.datetime.now)),
                ('date_valid_to', core.fields.DateTimeField(blank=True, db_column='DateValidTo', null=True)),
                ('replacement_uuid', models.UUIDField(db_column='ReplacementUUID', null=True)),
                ('thirdparty_id', models.CharField(db_column='ThirdpartyId', max_length=255, null=True)),
                ('code_tp', models.CharField(db_column='CodeTp', max_length=255, null=True)),
                ('code', models.CharField(db_column='Code', max_length=255)),
                ('code_ext', models.CharField(db_column='CodeExt', max_length=255, null=True)),
                ('date_due', core.fields.DateField(db_column='DateDue', null=True)),
                ('date_payed', core.fields.DateField(db_column='DatePayed', null=True)),
                ('amount_discount', models.DecimalField(db_column='AmountDiscount', decimal_places=2, default=0.0, max_digits=18, null=True)),
                ('amount_net', models.DecimalField(db_column='AmountNet', decimal_places=2, default=0.0, max_digits=18)),
                ('amount_total', models.DecimalField(db_column='AmountTotal', decimal_places=2, default=0.0, max_digits=18)),
                ('tax_analysis', models.JSONField(db_column='TaxAnalysis', null=True)),
                ('status', models.SmallIntegerField(choices=[(0, 'draft'), (1, 'validated'), (2, 'payed'), (3, 'cancelled'), (4, 'deleted'), (5, 'suspended')], db_column='Status', default=0)),
                ('currency_tp_code', models.CharField(db_column='CurrencyTpCode', default=invoice.models.get_default_currency, max_length=255)),
                ('currency_code', models.CharField(db_column='CurrencyCode', default=invoice.models.get_default_currency, max_length=255)),
                ('note', models.TextField(blank=True, db_column='Note', null=True)),
                ('terms', models.TextField(blank=True, db_column='Terms', null=True)),
                ('payment_reference', models.CharField(db_column='PaymentReference', max_length=255, null=True)),
                ('subject_id', models.CharField(db_column='SubjectId', max_length=255, null=True)),
                ('date_bill', core.fields.DateField(db_column='DateBill', default=datetime.date.today, null=True)),
                ('subject_type', models.ForeignKey(db_column='SubjectType', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='subject_type_bill', to='contenttypes.ContentType')),
                ('thirdparty_type', models.ForeignKey(db_column='ThirdpartyType', null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='contenttypes.ContentType')),
                ('user_created', models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='bill_user_created', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='bill_user_updated', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'tblBill',
                'managed': True,
            },
            bases=(invoice.mixins.GenericInvoiceQuerysetMixin, dirtyfields.dirtyfields.DirtyFieldsMixin, models.Model),
        ),
        migrations.CreateModel(
            name='BillEvent',
            fields=[
                ('id', models.UUIDField(db_column='UUID', default=None, editable=False, primary_key=True, serialize=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', models.JSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', core.fields.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', core.fields.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('message', models.CharField(db_column='Message', max_length=500, null=True)),
                ('event_type', models.SmallIntegerField(choices=[(0, 'message'), (1, 'status'), (2, 'warning'), (3, 'payment'), (4, 'payment_error')], db_column='Status', default=0)),
                ('bill', models.ForeignKey(db_column='BillId', on_delete=django.db.models.deletion.DO_NOTHING, related_name='events_bill', to='invoice.Bill')),
                ('user_created', models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='billevent_user_created', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='billevent_user_updated', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'tblBillEvent',
                'managed': True,
            },
            bases=(invoice.mixins.GenericInvoiceQuerysetMixin, dirtyfields.dirtyfields.DirtyFieldsMixin, models.Model),
        ),
        migrations.CreateModel(
            name='BillItem',
            fields=[
                ('id', models.UUIDField(db_column='UUID', default=None, editable=False, primary_key=True, serialize=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', models.JSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', core.fields.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', core.fields.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('date_valid_from', core.fields.DateTimeField(db_column='DateValidFrom', default=datetime.datetime.now)),
                ('date_valid_to', core.fields.DateTimeField(blank=True, db_column='DateValidTo', null=True)),
                ('replacement_uuid', models.UUIDField(db_column='ReplacementUUID', null=True)),
                ('code', models.CharField(db_column='Code', max_length=255)),
                ('description', models.TextField(blank=True, db_column='Description', null=True)),
                ('details', models.JSONField(db_column='Details', null=True)),
                ('ledger_account', models.CharField(db_column='LedgerAccount', max_length=255, null=True)),
                ('quantity', models.IntegerField(db_column='Quantity', default=0.0)),
                ('unit_price', models.DecimalField(db_column='UnitPrice', decimal_places=2, default=0.0, max_digits=18)),
                ('discount', models.DecimalField(db_column='Discount', decimal_places=2, default=0.0, max_digits=18)),
                ('tax_rate', models.UUIDField(db_column='CalculationUUID', null=True)),
                ('tax_analysis', models.JSONField(db_column='TaxAnalysis', null=True)),
                ('amount_total', models.DecimalField(db_column='AmountTotal', decimal_places=2, max_digits=18, null=True)),
                ('amount_net', models.DecimalField(db_column='AmountNet', decimal_places=2, default=0.0, max_digits=18)),
                ('line_id', models.CharField(db_column='LineId', max_length=255, null=True)),
                ('bill', models.ForeignKey(db_column='BillId', on_delete=django.db.models.deletion.DO_NOTHING, related_name='line_items_bill', to='invoice.Bill')),
                ('line_type', models.ForeignKey(db_column='LineType', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='line_type_bill', to='contenttypes.ContentType')),
                ('user_created', models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='billitem_user_created', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='billitem_user_updated', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'tblBillLineItem',
                'managed': True,
            },
            bases=(invoice.mixins.GenericInvoiceQuerysetMixin, dirtyfields.dirtyfields.DirtyFieldsMixin, models.Model),
        ),
        migrations.CreateModel(
            name='BillPayment',
            fields=[
                ('id', models.UUIDField(db_column='UUID', default=None, editable=False, primary_key=True, serialize=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', models.JSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', core.fields.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', core.fields.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('code_tp', models.CharField(db_column='CodeTp', max_length=255, null=True)),
                ('code_ext', models.CharField(db_column='CodeExt', max_length=255, null=True)),
                ('code_receipt', models.CharField(db_column='CodeReceipt', max_length=255, null=True)),
                ('label', models.CharField(db_column='Label', max_length=255, null=True)),
                ('status', models.SmallIntegerField(choices=[(0, 'rejected'), (1, 'accepted'), (2, 'refunded'), (3, 'cancelled')], db_column='Status')),
                ('amount_payed', models.DecimalField(db_column='AmountPayed', decimal_places=2, max_digits=18, null=True)),
                ('fees', models.DecimalField(db_column='Fees', decimal_places=2, max_digits=18, null=True)),
                ('amount_received', models.DecimalField(db_column='AmountReceived', decimal_places=2, max_digits=18, null=True)),
                ('date_payment', core.fields.DateField(db_column='DatePayment', null=True)),
                ('bill', models.ForeignKey(db_column='BillId', on_delete=django.db.models.deletion.DO_NOTHING, related_name='payments_bill', to='invoice.Bill')),
                ('user_created', models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='billpayment_user_created', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='billpayment_user_updated', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'tblBillPayment',
                'managed': True,
            },
            bases=(invoice.mixins.GenericInvoiceQuerysetMixin, dirtyfields.dirtyfields.DirtyFieldsMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Invoice',
            fields=[
                ('id', models.UUIDField(db_column='UUID', default=None, editable=False, primary_key=True, serialize=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', models.JSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', core.fields.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', core.fields.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('date_valid_from', core.fields.DateTimeField(db_column='DateValidFrom', default=datetime.datetime.now)),
                ('date_valid_to', core.fields.DateTimeField(blank=True, db_column='DateValidTo', null=True)),
                ('replacement_uuid', models.UUIDField(db_column='ReplacementUUID', null=True)),
                ('thirdparty_id', models.CharField(db_column='ThirdpartyId', max_length=255, null=True)),
                ('code_tp', models.CharField(db_column='CodeTp', max_length=255, null=True)),
                ('code', models.CharField(db_column='Code', max_length=255)),
                ('code_ext', models.CharField(db_column='CodeExt', max_length=255, null=True)),
                ('date_due', core.fields.DateField(db_column='DateDue', null=True)),
                ('date_payed', core.fields.DateField(db_column='DatePayed', null=True)),
                ('amount_discount', models.DecimalField(db_column='AmountDiscount', decimal_places=2, default=0.0, max_digits=18, null=True)),
                ('amount_net', models.DecimalField(db_column='AmountNet', decimal_places=2, default=0.0, max_digits=18)),
                ('amount_total', models.DecimalField(db_column='AmountTotal', decimal_places=2, default=0.0, max_digits=18)),
                ('tax_analysis', models.JSONField(db_column='TaxAnalysis', null=True)),
                ('status', models.SmallIntegerField(choices=[(0, 'draft'), (1, 'validated'), (2, 'payed'), (3, 'cancelled'), (4, 'deleted'), (5, 'suspended')], db_column='Status', default=0)),
                ('currency_tp_code', models.CharField(db_column='CurrencyTpCode', default=invoice.models.get_default_currency, max_length=255)),
                ('currency_code', models.CharField(db_column='CurrencyCode', default=invoice.models.get_default_currency, max_length=255)),
                ('note', models.TextField(blank=True, db_column='Note', null=True)),
                ('terms', models.TextField(blank=True, db_column='Terms', null=True)),
                ('payment_reference', models.CharField(db_column='PaymentReference', max_length=255, null=True)),
                ('subject_id', models.CharField(db_column='SubjectId', max_length=255, null=True)),
                ('date_invoice', core.fields.DateField(db_column='DateInvoice', default=datetime.date.today, null=True)),
                ('subject_type', models.ForeignKey(db_column='SubjectType', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='subject_type', to='contenttypes.ContentType')),
                ('thirdparty_type', models.ForeignKey(db_column='ThirdpartyType', null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='contenttypes.ContentType')),
                ('user_created', models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='invoice_user_created', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='invoice_user_updated', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'tblInvoice',
                'managed': True,
            },
            bases=(invoice.mixins.GenericInvoiceQuerysetMixin, dirtyfields.dirtyfields.DirtyFieldsMixin, models.Model),
        ),
        migrations.CreateModel(
            name='InvoiceEvent',
            fields=[
                ('id', models.UUIDField(db_column='UUID', default=None, editable=False, primary_key=True, serialize=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', models.JSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', core.fields.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', core.fields.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('message', models.CharField(db_column='Message', max_length=500, null=True)),
                ('event_type', models.SmallIntegerField(choices=[(0, 'message'), (1, 'status'), (2, 'warning'), (3, 'payment'), (4, 'payment_error')], db_column='Status', default=0)),
                ('invoice', models.ForeignKey(db_column='InvoiceId', on_delete=django.db.models.deletion.DO_NOTHING, related_name='events', to='invoice.Invoice')),
                ('user_created', models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='invoiceevent_user_created', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='invoiceevent_user_updated', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'tblInvoiceEvent',
                'managed': True,
            },
            bases=(invoice.mixins.GenericInvoiceQuerysetMixin, dirtyfields.dirtyfields.DirtyFieldsMixin, models.Model),
        ),
        migrations.CreateModel(
            name='InvoiceLineItem',
            fields=[
                ('id', models.UUIDField(db_column='UUID', default=None, editable=False, primary_key=True, serialize=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', models.JSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', core.fields.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', core.fields.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('date_valid_from', core.fields.DateTimeField(db_column='DateValidFrom', default=datetime.datetime.now)),
                ('date_valid_to', core.fields.DateTimeField(blank=True, db_column='DateValidTo', null=True)),
                ('replacement_uuid', models.UUIDField(db_column='ReplacementUUID', null=True)),
                ('code', models.CharField(db_column='Code', max_length=255)),
                ('description', models.TextField(blank=True, db_column='Description', null=True)),
                ('details', models.JSONField(db_column='Details', null=True)),
                ('ledger_account', models.CharField(db_column='LedgerAccount', max_length=255, null=True)),
                ('quantity', models.IntegerField(db_column='Quantity', default=0.0)),
                ('unit_price', models.DecimalField(db_column='UnitPrice', decimal_places=2, default=0.0, max_digits=18)),
                ('discount', models.DecimalField(db_column='Discount', decimal_places=2, default=0.0, max_digits=18)),
                ('tax_rate', models.UUIDField(db_column='CalculationUUID', null=True)),
                ('tax_analysis', models.JSONField(db_column='TaxAnalysis', null=True)),
                ('amount_total', models.DecimalField(db_column='AmountTotal', decimal_places=2, max_digits=18, null=True)),
                ('amount_net', models.DecimalField(db_column='AmountNet', decimal_places=2, default=0.0, max_digits=18)),
                ('line_id', models.CharField(db_column='LineId', max_length=255, null=True)),
                ('invoice', models.ForeignKey(db_column='InvoiceId', on_delete=django.db.models.deletion.DO_NOTHING, related_name='line_items', to='invoice.Invoice')),
                ('line_type', models.ForeignKey(db_column='LineType', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='line_type', to='contenttypes.ContentType')),
                ('user_created', models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='invoicelineitem_user_created', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='invoicelineitem_user_updated', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'tblInvoiceLineItem',
                'managed': True,
            },
            bases=(invoice.mixins.GenericInvoiceQuerysetMixin, dirtyfields.dirtyfields.DirtyFieldsMixin, models.Model),
        ),
        migrations.CreateModel(
            name='InvoicePayment',
            fields=[
                ('id', models.UUIDField(db_column='UUID', default=None, editable=False, primary_key=True, serialize=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', models.JSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', core.fields.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', core.fields.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('code_tp', models.CharField(db_column='CodeTp', max_length=255, null=True)),
                ('code_ext', models.CharField(db_column='CodeExt', max_length=255, null=True)),
                ('code_receipt', models.CharField(db_column='CodeReceipt', max_length=255, null=True)),
                ('label', models.CharField(db_column='Label', max_length=255, null=True)),
                ('status', models.SmallIntegerField(choices=[(0, 'rejected'), (1, 'accepted'), (2, 'refunded'), (3, 'cancelled')], db_column='Status')),
                ('amount_payed', models.DecimalField(db_column='AmountPayed', decimal_places=2, max_digits=18, null=True)),
                ('fees', models.DecimalField(db_column='Fees', decimal_places=2, max_digits=18, null=True)),
                ('amount_received', models.DecimalField(db_column='AmountReceived', decimal_places=2, max_digits=18, null=True)),
                ('date_payment', core.fields.DateField(db_column='DatePayment', null=True)),
                ('invoice', models.ForeignKey(db_column='InvoiceId', on_delete=django.db.models.deletion.DO_NOTHING, related_name='payments', to='invoice.Invoice')),
                ('user_created', models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='invoicepayment_user_created', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='invoicepayment_user_updated', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'tblInvoicePayment',
                'managed': True,
            },
            bases=(invoice.mixins.GenericInvoiceQuerysetMixin, dirtyfields.dirtyfields.DirtyFieldsMixin, models.Model),
        ),
        migrations.CreateModel(
            name='InvoicePaymentMutation',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('invoice_payment', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='mutations', to='invoice.InvoicePayment')),
                ('mutation', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='invoice_payments', to='core.MutationLog')),
            ],
            options={
                'db_table': 'invoice_InvoicePaymentMutation',
                'managed': True,
            },
            bases=(models.Model, core.models.ObjectMutation),
        ),
        migrations.CreateModel(
            name='InvoiceMutation',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('invoice', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='mutations', to='invoice.Invoice')),
                ('mutation', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='invoices', to='core.MutationLog')),
            ],
            options={
                'db_table': 'invoice_invoiceMutation',
                'managed': True,
            },
            bases=(models.Model, core.models.ObjectMutation),
        ),
        migrations.CreateModel(
            name='InvoiceLineItemMutation',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('invoice_line_items', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='mutations', to='invoice.InvoiceLineItem')),
                ('mutation', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='invoice_line_items', to='core.MutationLog')),
            ],
            options={
                'db_table': 'invoice_InvoiceLineItemsMutation',
                'managed': True,
            },
            bases=(models.Model, core.models.ObjectMutation),
        ),
        migrations.CreateModel(
            name='InvoiceEventMutation',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('invoice_event', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='mutations', to='invoice.InvoiceEvent')),
                ('mutation', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='event_messages', to='core.MutationLog')),
            ],
            options={
                'db_table': 'invoice_InvoiceEventMutation',
                'managed': True,
            },
            bases=(models.Model, core.models.ObjectMutation),
        ),
        migrations.CreateModel(
            name='HistoricalInvoicePayment',
            fields=[
                ('id', models.UUIDField(db_column='UUID', db_index=True, default=None, editable=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', models.JSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', core.fields.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', core.fields.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('code_tp', models.CharField(db_column='CodeTp', max_length=255, null=True)),
                ('code_ext', models.CharField(db_column='CodeExt', max_length=255, null=True)),
                ('code_receipt', models.CharField(db_column='CodeReceipt', max_length=255, null=True)),
                ('label', models.CharField(db_column='Label', max_length=255, null=True)),
                ('status', models.SmallIntegerField(choices=[(0, 'rejected'), (1, 'accepted'), (2, 'refunded'), (3, 'cancelled')], db_column='Status')),
                ('amount_payed', models.DecimalField(db_column='AmountPayed', decimal_places=2, max_digits=18, null=True)),
                ('fees', models.DecimalField(db_column='Fees', decimal_places=2, max_digits=18, null=True)),
                ('amount_received', models.DecimalField(db_column='AmountReceived', decimal_places=2, max_digits=18, null=True)),
                ('date_payment', core.fields.DateField(db_column='DatePayment', null=True)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('invoice', models.ForeignKey(blank=True, db_column='InvoiceId', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='invoice.Invoice')),
                ('user_created', models.ForeignKey(blank=True, db_column='UserCreatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(blank=True, db_column='UserUpdatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'historical invoice payment',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='HistoricalInvoiceLineItem',
            fields=[
                ('id', models.UUIDField(db_column='UUID', db_index=True, default=None, editable=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', models.JSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', core.fields.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', core.fields.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('date_valid_from', core.fields.DateTimeField(db_column='DateValidFrom', default=datetime.datetime.now)),
                ('date_valid_to', core.fields.DateTimeField(blank=True, db_column='DateValidTo', null=True)),
                ('replacement_uuid', models.UUIDField(db_column='ReplacementUUID', null=True)),
                ('code', models.CharField(db_column='Code', max_length=255)),
                ('description', models.TextField(blank=True, db_column='Description', null=True)),
                ('details', models.JSONField(db_column='Details', null=True)),
                ('ledger_account', models.CharField(db_column='LedgerAccount', max_length=255, null=True)),
                ('quantity', models.IntegerField(db_column='Quantity', default=0.0)),
                ('unit_price', models.DecimalField(db_column='UnitPrice', decimal_places=2, default=0.0, max_digits=18)),
                ('discount', models.DecimalField(db_column='Discount', decimal_places=2, default=0.0, max_digits=18)),
                ('tax_rate', models.UUIDField(db_column='CalculationUUID', null=True)),
                ('tax_analysis', models.JSONField(db_column='TaxAnalysis', null=True)),
                ('amount_total', models.DecimalField(db_column='AmountTotal', decimal_places=2, max_digits=18, null=True)),
                ('amount_net', models.DecimalField(db_column='AmountNet', decimal_places=2, default=0.0, max_digits=18)),
                ('line_id', models.CharField(db_column='LineId', max_length=255, null=True)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('invoice', models.ForeignKey(blank=True, db_column='InvoiceId', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='invoice.Invoice')),
                ('line_type', models.ForeignKey(blank=True, db_column='LineType', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='contenttypes.ContentType')),
                ('user_created', models.ForeignKey(blank=True, db_column='UserCreatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(blank=True, db_column='UserUpdatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'historical invoice line item',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='HistoricalInvoiceEvent',
            fields=[
                ('id', models.UUIDField(db_column='UUID', db_index=True, default=None, editable=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', models.JSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', core.fields.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', core.fields.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('message', models.CharField(db_column='Message', max_length=500, null=True)),
                ('event_type', models.SmallIntegerField(choices=[(0, 'message'), (1, 'status'), (2, 'warning'), (3, 'payment'), (4, 'payment_error')], db_column='Status', default=0)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('invoice', models.ForeignKey(blank=True, db_column='InvoiceId', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='invoice.Invoice')),
                ('user_created', models.ForeignKey(blank=True, db_column='UserCreatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(blank=True, db_column='UserUpdatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'historical invoice event',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='HistoricalInvoice',
            fields=[
                ('id', models.UUIDField(db_column='UUID', db_index=True, default=None, editable=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', models.JSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', core.fields.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', core.fields.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('date_valid_from', core.fields.DateTimeField(db_column='DateValidFrom', default=datetime.datetime.now)),
                ('date_valid_to', core.fields.DateTimeField(blank=True, db_column='DateValidTo', null=True)),
                ('replacement_uuid', models.UUIDField(db_column='ReplacementUUID', null=True)),
                ('thirdparty_id', models.CharField(db_column='ThirdpartyId', max_length=255, null=True)),
                ('code_tp', models.CharField(db_column='CodeTp', max_length=255, null=True)),
                ('code', models.CharField(db_column='Code', max_length=255)),
                ('code_ext', models.CharField(db_column='CodeExt', max_length=255, null=True)),
                ('date_due', core.fields.DateField(db_column='DateDue', null=True)),
                ('date_payed', core.fields.DateField(db_column='DatePayed', null=True)),
                ('amount_discount', models.DecimalField(db_column='AmountDiscount', decimal_places=2, default=0.0, max_digits=18, null=True)),
                ('amount_net', models.DecimalField(db_column='AmountNet', decimal_places=2, default=0.0, max_digits=18)),
                ('amount_total', models.DecimalField(db_column='AmountTotal', decimal_places=2, default=0.0, max_digits=18)),
                ('tax_analysis', models.JSONField(db_column='TaxAnalysis', null=True)),
                ('status', models.SmallIntegerField(choices=[(0, 'draft'), (1, 'validated'), (2, 'payed'), (3, 'cancelled'), (4, 'deleted'), (5, 'suspended')], db_column='Status', default=0)),
                ('currency_tp_code', models.CharField(db_column='CurrencyTpCode', default=invoice.models.get_default_currency, max_length=255)),
                ('currency_code', models.CharField(db_column='CurrencyCode', default=invoice.models.get_default_currency, max_length=255)),
                ('note', models.TextField(blank=True, db_column='Note', null=True)),
                ('terms', models.TextField(blank=True, db_column='Terms', null=True)),
                ('payment_reference', models.CharField(db_column='PaymentReference', max_length=255, null=True)),
                ('subject_id', models.CharField(db_column='SubjectId', max_length=255, null=True)),
                ('date_invoice', core.fields.DateField(db_column='DateInvoice', default=datetime.date.today, null=True)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('subject_type', models.ForeignKey(blank=True, db_column='SubjectType', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='contenttypes.ContentType')),
                ('thirdparty_type', models.ForeignKey(blank=True, db_column='ThirdpartyType', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='contenttypes.ContentType')),
                ('user_created', models.ForeignKey(blank=True, db_column='UserCreatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(blank=True, db_column='UserUpdatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'historical invoice',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='HistoricalBillPayment',
            fields=[
                ('id', models.UUIDField(db_column='UUID', db_index=True, default=None, editable=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', models.JSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', core.fields.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', core.fields.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('code_tp', models.CharField(db_column='CodeTp', max_length=255, null=True)),
                ('code_ext', models.CharField(db_column='CodeExt', max_length=255, null=True)),
                ('code_receipt', models.CharField(db_column='CodeReceipt', max_length=255, null=True)),
                ('label', models.CharField(db_column='Label', max_length=255, null=True)),
                ('status', models.SmallIntegerField(choices=[(0, 'rejected'), (1, 'accepted'), (2, 'refunded'), (3, 'cancelled')], db_column='Status')),
                ('amount_payed', models.DecimalField(db_column='AmountPayed', decimal_places=2, max_digits=18, null=True)),
                ('fees', models.DecimalField(db_column='Fees', decimal_places=2, max_digits=18, null=True)),
                ('amount_received', models.DecimalField(db_column='AmountReceived', decimal_places=2, max_digits=18, null=True)),
                ('date_payment', core.fields.DateField(db_column='DatePayment', null=True)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('bill', models.ForeignKey(blank=True, db_column='BillId', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='invoice.Bill')),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('user_created', models.ForeignKey(blank=True, db_column='UserCreatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(blank=True, db_column='UserUpdatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'historical bill payment',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='HistoricalBillItem',
            fields=[
                ('id', models.UUIDField(db_column='UUID', db_index=True, default=None, editable=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', models.JSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', core.fields.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', core.fields.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('date_valid_from', core.fields.DateTimeField(db_column='DateValidFrom', default=datetime.datetime.now)),
                ('date_valid_to', core.fields.DateTimeField(blank=True, db_column='DateValidTo', null=True)),
                ('replacement_uuid', models.UUIDField(db_column='ReplacementUUID', null=True)),
                ('code', models.CharField(db_column='Code', max_length=255)),
                ('description', models.TextField(blank=True, db_column='Description', null=True)),
                ('details', models.JSONField(db_column='Details', null=True)),
                ('ledger_account', models.CharField(db_column='LedgerAccount', max_length=255, null=True)),
                ('quantity', models.IntegerField(db_column='Quantity', default=0.0)),
                ('unit_price', models.DecimalField(db_column='UnitPrice', decimal_places=2, default=0.0, max_digits=18)),
                ('discount', models.DecimalField(db_column='Discount', decimal_places=2, default=0.0, max_digits=18)),
                ('tax_rate', models.UUIDField(db_column='CalculationUUID', null=True)),
                ('tax_analysis', models.JSONField(db_column='TaxAnalysis', null=True)),
                ('amount_total', models.DecimalField(db_column='AmountTotal', decimal_places=2, max_digits=18, null=True)),
                ('amount_net', models.DecimalField(db_column='AmountNet', decimal_places=2, default=0.0, max_digits=18)),
                ('line_id', models.CharField(db_column='LineId', max_length=255, null=True)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('bill', models.ForeignKey(blank=True, db_column='BillId', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='invoice.Bill')),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('line_type', models.ForeignKey(blank=True, db_column='LineType', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='contenttypes.ContentType')),
                ('user_created', models.ForeignKey(blank=True, db_column='UserCreatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(blank=True, db_column='UserUpdatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'historical bill item',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='HistoricalBillEvent',
            fields=[
                ('id', models.UUIDField(db_column='UUID', db_index=True, default=None, editable=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', models.JSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', core.fields.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', core.fields.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('message', models.CharField(db_column='Message', max_length=500, null=True)),
                ('event_type', models.SmallIntegerField(choices=[(0, 'message'), (1, 'status'), (2, 'warning'), (3, 'payment'), (4, 'payment_error')], db_column='Status', default=0)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('bill', models.ForeignKey(blank=True, db_column='BillId', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='invoice.Bill')),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('user_created', models.ForeignKey(blank=True, db_column='UserCreatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(blank=True, db_column='UserUpdatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'historical bill event',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='HistoricalBill',
            fields=[
                ('id', models.UUIDField(db_column='UUID', db_index=True, default=None, editable=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', models.JSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', core.fields.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', core.fields.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('date_valid_from', core.fields.DateTimeField(db_column='DateValidFrom', default=datetime.datetime.now)),
                ('date_valid_to', core.fields.DateTimeField(blank=True, db_column='DateValidTo', null=True)),
                ('replacement_uuid', models.UUIDField(db_column='ReplacementUUID', null=True)),
                ('thirdparty_id', models.CharField(db_column='ThirdpartyId', max_length=255, null=True)),
                ('code_tp', models.CharField(db_column='CodeTp', max_length=255, null=True)),
                ('code', models.CharField(db_column='Code', max_length=255)),
                ('code_ext', models.CharField(db_column='CodeExt', max_length=255, null=True)),
                ('date_due', core.fields.DateField(db_column='DateDue', null=True)),
                ('date_payed', core.fields.DateField(db_column='DatePayed', null=True)),
                ('amount_discount', models.DecimalField(db_column='AmountDiscount', decimal_places=2, default=0.0, max_digits=18, null=True)),
                ('amount_net', models.DecimalField(db_column='AmountNet', decimal_places=2, default=0.0, max_digits=18)),
                ('amount_total', models.DecimalField(db_column='AmountTotal', decimal_places=2, default=0.0, max_digits=18)),
                ('tax_analysis', models.JSONField(db_column='TaxAnalysis', null=True)),
                ('status', models.SmallIntegerField(choices=[(0, 'draft'), (1, 'validated'), (2, 'payed'), (3, 'cancelled'), (4, 'deleted'), (5, 'suspended')], db_column='Status', default=0)),
                ('currency_tp_code', models.CharField(db_column='CurrencyTpCode', default=invoice.models.get_default_currency, max_length=255)),
                ('currency_code', models.CharField(db_column='CurrencyCode', default=invoice.models.get_default_currency, max_length=255)),
                ('note', models.TextField(blank=True, db_column='Note', null=True)),
                ('terms', models.TextField(blank=True, db_column='Terms', null=True)),
                ('payment_reference', models.CharField(db_column='PaymentReference', max_length=255, null=True)),
                ('subject_id', models.CharField(db_column='SubjectId', max_length=255, null=True)),
                ('date_bill', core.fields.DateField(db_column='DateBill', default=datetime.date.today, null=True)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('subject_type', models.ForeignKey(blank=True, db_column='SubjectType', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='contenttypes.ContentType')),
                ('thirdparty_type', models.ForeignKey(blank=True, db_column='ThirdpartyType', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='contenttypes.ContentType')),
                ('user_created', models.ForeignKey(blank=True, db_column='UserCreatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(blank=True, db_column='UserUpdatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'historical bill',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='BillPaymentMutation',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('bill_payment', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='mutations', to='invoice.BillPayment')),
                ('mutation', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='bill_payments', to='core.MutationLog')),
            ],
            options={
                'db_table': 'bill_BillPaymentMutation',
                'managed': True,
            },
            bases=(models.Model, core.models.ObjectMutation),
        ),
        migrations.CreateModel(
            name='BillMutation',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('bill', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='mutations', to='invoice.Bill')),
                ('mutation', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='bills', to='core.MutationLog')),
            ],
            options={
                'db_table': 'bill_BillMutation',
                'managed': True,
            },
            bases=(models.Model, core.models.ObjectMutation),
        ),
        migrations.CreateModel(
            name='BillItemMutation',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('bile_items', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='mutations', to='invoice.BillItem')),
                ('mutation', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='bill_line_items', to='core.MutationLog')),
            ],
            options={
                'db_table': 'bill_BillLineItemsMutation',
                'managed': True,
            },
            bases=(models.Model, core.models.ObjectMutation),
        ),
        migrations.CreateModel(
            name='BillEventMutation',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('bill_event', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='mutations', to='invoice.BillEvent')),
                ('mutation', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='bill_event_messages', to='core.MutationLog')),
            ],
            options={
                'db_table': 'bill_BillEventMutation',
                'managed': True,
            },
            bases=(models.Model, core.models.ObjectMutation),
        ),
    ]



from django.contrib.contenttypes.models import ContentType

from core.forms import User
from invoice.models import PaymentInvoice, DetailPaymentInvoice
from invoice.tests.helpers import create_test_invoice, create_test_invoice_line_item
from invoice.tests.helpers.default_test_data import DEFAULT_TEST_PAYMENT_INVOICE_PAYLOAD, \
    DEFAULT_TEST_DETAIL_PAYMENT_INVOICE_PAYLOAD


def create_test_payment_invoice_without_details(invoice=None, user=None, **custom_props):
    invoice = invoice or create_test_invoice()
    invoice_item = create_test_invoice_line_item(invoice=invoice)
    user = user or __get_or_create_user()

    payload = DEFAULT_TEST_PAYMENT_INVOICE_PAYLOAD.copy()
    payload.update(**custom_props)

    payment = PaymentInvoice(**payload)
    payment.save(username=user.username)

    return payment, invoice, invoice_item


def create_test_payment_invoice_with_details(invoice=None, user=None, **custom_props):
    invoice = invoice or create_test_invoice()
    create_test_invoice_line_item(invoice=invoice)
    user = user or __get_or_create_user()

    payload = DEFAULT_TEST_PAYMENT_INVOICE_PAYLOAD.copy()
    payload.update(**custom_props)

    payment = PaymentInvoice(**payload)
    payment.save(username=user.username)

    payload_detail = DEFAULT_TEST_DETAIL_PAYMENT_INVOICE_PAYLOAD.copy()
    payload_detail['subject'] = invoice
    payload_detail['subject_type'] = ContentType.objects.get_for_model(invoice)
    payload_detail.update(**custom_props)
    payment_detail = DetailPaymentInvoice(**payload_detail)
    payment_detail.payment = payment
    payment_detail.save(username=user.username)
    return payment


def __get_or_create_user():
    user = User.objects.filter(username='admin_invoice').first()
    if not user:
        user = User.objects.create_superuser(username='admin_invoice', password='S\/pe®Pąßw0rd™')
    return user



import json
import graphene

from django.core.serializers.json import DjangoJSONEncoder
from graphene_django import DjangoObjectType

from core import prefix_filterset, ExtendedConnection
from invoice.apps import InvoiceConfig
from invoice.gql.filter_mixin import GenericFilterGQLTypeMixin
from invoice.models import PaymentInvoice, DetailPaymentInvoice
from invoice.utils import underscore_to_camel
from django.utils.translation import gettext as _
from django.core.exceptions import PermissionDenied


class PaymentInvoiceGQLType(DjangoObjectType, GenericFilterGQLTypeMixin):

    class Meta:
        model = PaymentInvoice
        interfaces = (graphene.relay.Node,)
        filter_fields = {
            **GenericFilterGQLTypeMixin.get_base_filters_payment_invoice(),
        }

        connection_class = ExtendedConnection

        @classmethod
        def get_queryset(cls, queryset, info):
            return PaymentInvoice.get_queryset(queryset, info)


class DetailPaymentInvoiceGQLType(DjangoObjectType, GenericFilterGQLTypeMixin):

    subject_type = graphene.Int()
    def resolve_subject_type(root, info):
        if not info.context.user.has_perms(InvoiceConfig.gql_invoice_payment_search_perms):
            raise PermissionDenied(_("unauthorized"))
        return root.subject_type.id

    subject_type_name = graphene.String()
    def resolve_subject_type_name(root, info):
        if not info.context.user.has_perms(InvoiceConfig.gql_invoice_payment_search_perms):
            raise PermissionDenied(_("unauthorized"))
        return root.subject_type.name

    subject = graphene.JSONString()
    def resolve_subject(root, info):
        if not info.context.user.has_perms(InvoiceConfig.gql_invoice_payment_search_perms):
            raise PermissionDenied(_("unauthorized"))
        subject_object_dict = root.subject.__dict__
        subject_object_dict.pop('_state')
        subject_object_dict = {
            underscore_to_camel(k): v for k, v in list(subject_object_dict.items())
        }
        subject_object_dict = json.dumps(subject_object_dict, cls=DjangoJSONEncoder)
        return subject_object_dict

    class Meta:
        model = DetailPaymentInvoice
        interfaces = (graphene.relay.Node,)
        filter_fields = {
            **GenericFilterGQLTypeMixin.get_base_filters_detail_invoice_payment(),
            **prefix_filterset("payment__", PaymentInvoiceGQLType._meta.filter_fields),
        }

        connection_class = ExtendedConnection

        @classmethod
        def get_queryset(cls, queryset, info):
            return DetailPaymentInvoice.get_queryset(queryset, info)



import copy

from django.db import transaction
from unittest import skip
from policy.test_helpers import create_test_policy
from product.test_helpers import create_test_product

from core.forms import User
from django.test import TestCase

from invoice.models import Invoice, InvoicePayment
from contract.tests.helpers import create_test_contract
from policyholder.tests.helpers import create_test_policy_holder
from insuree.test_helpers import create_test_insuree
from datetime import date

from invoice.services.invoicePayments import InvoicePaymentsService
from invoice.tests.helpers import create_test_invoice, create_test_invoice_line_item
from invoice.validation.invoicePayment import InvoicePaymentModelValidation
from invoice.validation.paymentStatusValidation import InvoicePaymentRefundStatusValidator, \
    InvoicePaymentReceiveStatusValidator, InvoicePaymentCancelStatusValidator


class ServiceTestInvoicePayments(TestCase):
    BASE_TEST_INVOICE_PAYMENT_PAYLOAD = {
        'label': 'label_pay',
        'code_tp': 'pay_sys_ref',
        'code_receipt': 'receipt number',
        'invoice': None,
        'amount_payed': 91.5,
        'fees': 12.0,
        'amount_received': 22.0,
        'date_payment': date(2021, 10, 10),
    }

    BASE_EXPECTED_SUCCESS_RESPONSE = {
        "success": True,
        "message": "Ok",
        "detail": "",
        "data": {
            'status': 1,
            'label': 'label_pay',
            'code_tp': 'pay_sys_ref',
            'code_receipt': 'receipt number',
            'amount_payed': 91.5,
            'fees': 12.0,
            'amount_received': 22.00,
            'date_payment': str(date(2021, 10, 10))
        },
    }

    @classmethod
    def setUpClass(cls):
        super(ServiceTestInvoicePayments, cls).setUpClass()
        cls.maxDiff = None
        if not User.objects.filter(username='admin_invoice').exists():
            User.objects.create_superuser(username='admin_invoice', password='S\/pe®Pąßw0rd™')

        cls.user = User.objects.filter(username='admin_invoice').first()
        cls.invoice_payment_service = InvoicePaymentsService(cls.user)

        cls.policy_holder = create_test_policy_holder()
        cls.contract = create_test_contract(cls.policy_holder)
        cls.insuree = create_test_insuree(with_family=True)
        cls.product = create_test_product("TestC0d3", custom_props={"insurance_period": 12})
        cls.policy = create_test_policy(
            product=cls.product,
            insuree=cls.insuree
        )

        cls.invoice = create_test_invoice(cls.contract, cls.insuree)
        cls.invoice_line_item = \
            create_test_invoice_line_item(invoice=cls.invoice, line_item=cls.policy, user=cls.user)

    @skip("This is related to old payment approach model - depreceated")
    def test_ref_received(self):
        with transaction.atomic():
            payload = self.BASE_TEST_INVOICE_PAYMENT_PAYLOAD.copy()
            payload['status'] = InvoicePayment.PaymentStatus.ACCEPTED

            payment = self._create_payment(payload)
            payment.save(username=self.user.username)

            out = self.invoice_payment_service.ref_received(payment, 'code_ext1')
            expected = self.BASE_EXPECTED_SUCCESS_RESPONSE.copy()
            expected['data']['code_ext'] = 'code_ext1'
            self._assert_output_valid(out, payment, expected)
            InvoicePayment.objects.filter(code_ext=payment.code_ext).delete()

    @skip("This is related to old payment approach model - depreceated")
    def test_payment_received(self):
        with transaction.atomic():
            payment = self._create_payment(self.BASE_TEST_INVOICE_PAYMENT_PAYLOAD)
            out = self.invoice_payment_service.payment_received(payment, InvoicePayment.PaymentStatus.ACCEPTED)
            self._assert_output_valid(out, payment, self.BASE_EXPECTED_SUCCESS_RESPONSE.copy())
            InvoicePayment.objects.filter(code_ext=payment.code_ext).delete()

    @skip("This is related to old payment approach model - depreceated")
    def test_payment_received_invalid_amount(self):
        with transaction.atomic():
            payload = self.BASE_TEST_INVOICE_PAYMENT_PAYLOAD.copy()
            payload['amount_payed'] = 2.0
            payment = self._create_payment(payload)
            out = self.invoice_payment_service.payment_received(payment, InvoicePayment.PaymentStatus.ACCEPTED)
            detail = [str(InvoicePaymentModelValidation.AMOUNT_PAYED_NOT_MATCHING_ITEMS \
                          % {'invoice': self.invoice, 'expected': 91.5, 'payed': 2.0})]
            message = 'Failed to payment_received InvoicePayment'
            self._assert_output_invalid(out, message, detail)
            InvoicePayment.objects.filter(code_ext=payment.code_ext).delete()

    @skip("This is related to old payment approach model - depreceated")
    def test_payment_received_invalid_invoice_status(self):
        with transaction.atomic():
            payload = self.BASE_TEST_INVOICE_PAYMENT_PAYLOAD.copy()
            payment = self._create_payment(payload)
            payment.invoice.status = Invoice.Status.CANCELLED
            out = self.invoice_payment_service.payment_received(payment, InvoicePayment.PaymentStatus.ACCEPTED)
            detail = [str(InvoicePaymentReceiveStatusValidator.error_message_invalid_invoice \
                          % {'invoice': self.invoice, 'allowed_invoice': 'draft, validated',
                             'invoice_status': self.invoice.status.label})]
            message = 'Failed to payment_received InvoicePayment'
            self._assert_output_invalid(out, message, detail)

            payment.invoice.status = Invoice.Status.VALIDATED
            InvoicePayment.objects.filter(code_ext=payment.code_ext).delete()

    @skip("This is related to old payment approach model - depreceated")
    def test_payment_received_invalid_payment_status(self):
        with transaction.atomic():
            payload = self.BASE_TEST_INVOICE_PAYMENT_PAYLOAD.copy()
            payload['status'] = InvoicePayment.PaymentStatus.REFUNDED
            payment = self._create_payment(payload)
            out = self.invoice_payment_service.payment_received(payment, InvoicePayment.PaymentStatus.REFUNDED)
            detail = [str(InvoicePaymentReceiveStatusValidator.error_message_invalid_payment
                          % {'payment_status': InvoicePayment.PaymentStatus.REFUNDED.label,
                             'allowed_payment': 'accepted, rejected'})]
            message = 'Failed to payment_received InvoicePayment'
            self._assert_output_invalid(out, message, detail)
            InvoicePayment.objects.filter(code_ext=payment.code_ext).delete()

    @skip("This is related to old payment approach model - depreceated")
    def test_payment_refunded(self):
        with transaction.atomic():
            payment = self._create_payment(self.BASE_TEST_INVOICE_PAYMENT_PAYLOAD)
            payment.invoice.status = Invoice.Status.PAID
            payment.status = InvoicePayment.PaymentStatus.ACCEPTED

            out = self.invoice_payment_service.payment_refunded(payment)
            expected = copy.deepcopy(self.BASE_EXPECTED_SUCCESS_RESPONSE)
            expected['data']['status'] = 2
            self._assert_output_valid(out, payment, expected)
            self.assertEqual(payment.invoice.status, Invoice.Status.SUSPENDED)
            InvoicePayment.objects.filter(code_ext=payment.code_ext).delete()
            self.invoice.status = Invoice.Status.VALIDATED

    @skip("This is related to old payment approach model - depreceated")
    def test_payment_refunded_invalid_invoice_status(self):
        with transaction.atomic():
            payment = self._create_payment(self.BASE_TEST_INVOICE_PAYMENT_PAYLOAD)
            payment.invoice.status = Invoice.Status.CANCELLED
            payment.status = InvoicePayment.PaymentStatus.ACCEPTED

            out = self.invoice_payment_service.payment_refunded(payment)
            detail = [str(InvoicePaymentRefundStatusValidator.error_message_invalid_invoice
                          % {'invoice': self.invoice,
                             'invoice_status': self.invoice.status.label,
                             'allowed_invoice': Invoice.Status.PAID.label})]
            message = 'Failed to payment_refunded InvoicePayment'
            self._assert_output_invalid(out, message, detail)
            self.invoice.status = Invoice.Status.VALIDATED

    @skip("This is related to old payment approach model - depreceated")
    def test_payment_refunded_invalid_payment_status(self):
        with transaction.atomic():
            payment = self._create_payment(self.BASE_TEST_INVOICE_PAYMENT_PAYLOAD)
            payment.invoice.status = Invoice.Status.PAID
            payment.status = InvoicePayment.PaymentStatus.REJECTED

            out = self.invoice_payment_service.payment_refunded(payment)
            detail = [str(InvoicePaymentRefundStatusValidator.error_message_invalid_payment
                          % {'payment_status': payment.status.label, 'payment': payment})]
            message = 'Failed to payment_refunded InvoicePayment'
            self._assert_output_invalid(out, message, detail)
            self.invoice.status = Invoice.Status.VALIDATED

    @skip("This is related to old payment approach model - depreceated")
    def test_payment_cancel_invalid_invoice_status(self):
        with transaction.atomic():
            payment = self._create_payment(self.BASE_TEST_INVOICE_PAYMENT_PAYLOAD)
            payment.invoice.status = Invoice.Status.CANCELLED
            payment.status = InvoicePayment.PaymentStatus.ACCEPTED

            out = self.invoice_payment_service.payment_cancelled(payment)
            detail = [str(InvoicePaymentCancelStatusValidator.error_message_invalid_invoice
                          % {'invoice': self.invoice,
                             'invoice_status': self.invoice.status.label,
                             'allowed_invoice': Invoice.Status.PAID.label})]
            message = 'Failed to payment_refunded InvoicePayment'
            self._assert_output_invalid(out, message, detail)
            self.invoice.status = Invoice.Status.VALIDATED

    @skip("This is related to old payment approach model - depreceated")
    def test_payment_cancel_invalid_payment_status(self):
        with transaction.atomic():
            payment = self._create_payment(self.BASE_TEST_INVOICE_PAYMENT_PAYLOAD)
            payment.invoice.status = Invoice.Status.PAID
            payment.status = InvoicePayment.PaymentStatus.REJECTED

            out = self.invoice_payment_service.payment_cancelled(payment)
            detail = [str(InvoicePaymentCancelStatusValidator.error_message_invalid_payment
                          % {'payment_status': payment.status.label, 'payment': payment})]
            message = 'Failed to payment_refunded InvoicePayment'
            self._assert_output_invalid(out, message, detail)
            self.invoice.status = Invoice.Status.VALIDATED

    @skip("This is related to old payment approach model - depreceated")
    def test_payment_cancelled(self):
        with transaction.atomic():
            payment = self._create_payment(self.BASE_TEST_INVOICE_PAYMENT_PAYLOAD)
            payment.invoice.status = Invoice.Status.PAID
            payment.status = InvoicePayment.PaymentStatus.ACCEPTED

            out = self.invoice_payment_service.payment_cancelled(payment)
            expected = copy.deepcopy(self.BASE_EXPECTED_SUCCESS_RESPONSE)
            self._assert_output_valid(out, payment, expected)
            InvoicePayment.objects.filter(code_ext=payment.code_ext).delete()
            self.assertEqual(payment.invoice.status, Invoice.Status.SUSPENDED)
            self.invoice.status = Invoice.Status.VALIDATED

    def _create_payment(self, args):
        payment = InvoicePayment(**args)
        payment.invoice = self.invoice
        return payment

    def _assert_output_valid(self, out, payment, expected):
        expected['data']['invoice'] = str(self.invoice.uuid)
        expected['data']['id'] = str(payment.id)
        out['data'] = {k: v for k, v in out['data'].items() if k in expected['data'].keys()}
        self.assertDictEqual(out, expected)

    def _assert_output_invalid(self, out, message, detail):
        expected = {
            'success': False,
            'message': message,
            'detail': str(detail).replace('\'', '').replace('\"', ''),
            'data': ''
        }
        out['detail'] = out['detail'].replace('\'', '').replace('\"', '')
        self.assertDictEqual(out, expected)



from datetime import date
from django.contrib.contenttypes.models import ContentType
from django.db import transaction
from django.test import TestCase

from contract.tests.helpers import create_test_contract
from core.forms import User
from insuree.test_helpers import create_test_insuree
from invoice.models import Invoice, PaymentInvoice, DetailPaymentInvoice
from invoice.services.paymentInvoice import PaymentInvoiceService
from invoice.tests.helpers import (
    create_test_invoice,
    create_test_invoice_line_item,
    DEFAULT_TEST_PAYMENT_INVOICE_PAYLOAD,
    DEFAULT_TEST_DETAIL_PAYMENT_INVOICE_PAYLOAD
)
from invoice.validation.paymentInvoice import PaymentInvoiceModelValidation
from product.test_helpers import create_test_product
from policy.test_helpers import create_test_policy
from policyholder.tests.helpers import create_test_policy_holder


class ServiceTestPaymentInvoice(TestCase):
    BASE_EXPECTED_SUCCESS_RESPONSE = {
        "success": True,
        "message": "Ok",
        "detail": "",
        "data": {
            'status': 1,
            'label': 'test label',
            'code_tp': 'PAY_CODE',
            'code_receipt': 'PAY_CODE_RCP',
            'fees': 12.0,
            'amount_received': 91.50,
            'date_payment': str(date(2022, 4, 11))
        },
    }

    @classmethod
    def setUpClass(cls):
        super(ServiceTestPaymentInvoice, cls).setUpClass()
        cls.maxDiff = None
        if not User.objects.filter(username='admin_invoice').exists():
            User.objects.create_superuser(username='admin_invoice', password='S\/pe®Pąßw0rd™')

        cls.user = User.objects.filter(username='admin_invoice').first()
        cls.payment_invoice_service = PaymentInvoiceService(cls.user)

        cls.policy_holder = create_test_policy_holder()
        cls.contract = create_test_contract(cls.policy_holder)
        cls.insuree = create_test_insuree(with_family=True)
        cls.product = create_test_product("TestC0d3", custom_props={"insurance_period": 12})
        cls.policy = create_test_policy(
            product=cls.product,
            insuree=cls.insuree
        )

        cls.invoice = create_test_invoice(cls.contract, cls.insuree)
        cls.invoice_line_item = \
            create_test_invoice_line_item(invoice=cls.invoice, line_item=cls.policy, user=cls.user)

    def test_ref_received(self):
        with transaction.atomic():
            payment, payment_detail = self._create_payment(
                DEFAULT_TEST_PAYMENT_INVOICE_PAYLOAD,
                DEFAULT_TEST_DETAIL_PAYMENT_INVOICE_PAYLOAD
            )
            out = self.payment_invoice_service.ref_received(payment, 'code_ext1')
            expected = self.BASE_EXPECTED_SUCCESS_RESPONSE.copy()
            expected['data']['code_ext'] = 'code_ext1'
            payment_invoice = PaymentInvoice.objects.filter(code_ext=payment.code_ext).first()
            self.assertEqual(payment_invoice.code_ext, expected['data']['code_ext'])
            DetailPaymentInvoice.objects.filter(payment__code_ext=payment.code_ext).delete()
            PaymentInvoice.objects.filter(code_ext=payment.code_ext).delete()

    def test_payment_received(self):
        with transaction.atomic():
            payment, payment_detail = self._create_payment(
                DEFAULT_TEST_PAYMENT_INVOICE_PAYLOAD,
                DEFAULT_TEST_DETAIL_PAYMENT_INVOICE_PAYLOAD
            )
            self.payment_invoice_service.payment_received(
                payment,
                DetailPaymentInvoice.DetailPaymentStatus.ACCEPTED
            )
            detail_payment_invoice = DetailPaymentInvoice.objects.filter(payment__code_ext=payment.code_ext).first()
            self.assertEqual(detail_payment_invoice.subject.status, Invoice.Status.PAID)
            self.assertEqual(detail_payment_invoice.status, DetailPaymentInvoice.DetailPaymentStatus.ACCEPTED)
            DetailPaymentInvoice.objects.filter(payment__code_ext=payment.code_ext).delete()
            PaymentInvoice.objects.filter(code_ext=payment.code_ext).delete()

    def test_payment_received_invalid_amount(self):
        with transaction.atomic():
            payload = DEFAULT_TEST_PAYMENT_INVOICE_PAYLOAD.copy()
            payload['amount_received'] = 2.0
            payment, payment_detail = self._create_payment(
                payload,
                DEFAULT_TEST_DETAIL_PAYMENT_INVOICE_PAYLOAD
            )
            out = self.payment_invoice_service.payment_received(
                payment,
                DetailPaymentInvoice.DetailPaymentStatus.ACCEPTED
            )
            detail = [str(PaymentInvoiceModelValidation.AMOUNT_PAYED_NOT_MATCHING_ITEMS \
                          % {'payment_invoice': payment, 'expected': 91.5, 'payed': 2.0})]
            message = 'Failed to payment_received PaymentInvoice'
            self._assert_output_invalid(out, message, detail)
            DetailPaymentInvoice.objects.filter(payment__code_ext=payment.code_ext).delete()
            PaymentInvoice.objects.filter(code_ext=payment.code_ext).delete()

    def test_payment_refunded(self):
        with transaction.atomic():
            payment, payment_detail = self._create_payment(
                DEFAULT_TEST_PAYMENT_INVOICE_PAYLOAD,
                DEFAULT_TEST_DETAIL_PAYMENT_INVOICE_PAYLOAD
            )
            self.payment_invoice_service.payment_refunded(payment)
            detail_payment_invoice = DetailPaymentInvoice.objects.filter(payment__code_ext=payment.code_ext).first()
            self.assertEqual(detail_payment_invoice.subject.status, Invoice.Status.SUSPENDED)
            self.assertEqual(detail_payment_invoice.status, DetailPaymentInvoice.DetailPaymentStatus.REFUNDED)
            DetailPaymentInvoice.objects.filter(payment__code_ext=payment.code_ext).delete()
            PaymentInvoice.objects.filter(code_ext=payment.code_ext).delete()

    def test_payment_cancelled(self):
        with transaction.atomic():
            payment, payment_detail = self._create_payment(
                DEFAULT_TEST_PAYMENT_INVOICE_PAYLOAD,
                DEFAULT_TEST_DETAIL_PAYMENT_INVOICE_PAYLOAD
            )
            self.payment_invoice_service.payment_cancelled(payment)
            detail_payment_invoice = DetailPaymentInvoice.objects.filter(payment__code_ext=payment.code_ext).first()
            self.assertEqual(detail_payment_invoice.subject.status, Invoice.Status.SUSPENDED)
            self.assertEqual(detail_payment_invoice.status, DetailPaymentInvoice.DetailPaymentStatus.CANCELLED)
            DetailPaymentInvoice.objects.filter(payment__code_ext=payment.code_ext).delete()
            PaymentInvoice.objects.filter(code_ext=payment.code_ext).delete()

    def _create_payment(self, payment, payment_detail):
        payment = PaymentInvoice(**payment)
        payment.invoice = self.invoice
        payment.save(username=self.user.username)
        payload_detail = DEFAULT_TEST_DETAIL_PAYMENT_INVOICE_PAYLOAD.copy()
        payload_detail['subject'] = self.invoice
        payload_detail['subject_type'] = ContentType.objects.get_for_model(self.invoice)
        payment_detail = DetailPaymentInvoice(**payload_detail)
        payment_detail.payment = payment
        payment_detail.save(username=self.user.username)
        return payment, payment_detail

    def _assert_output_valid(self, out, payment, expected):
        expected['data']['id'] = str(payment.id)
        out['data'] = {k: v for k, v in out['data'].items() if k in expected['data'].keys()}
        self.assertDictEqual(out, expected)

    def _assert_output_invalid(self, out, message, detail):
        expected = {
            'success': False,
            'message': message,
            'detail': str(detail).replace('\'', '').replace('\"', ''),
            'data': ''
        }
        out['detail'] = out['detail'].replace('\'', '').replace('\"', '')
        self.assertDictEqual(out, expected)



from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _

from core.validation import BaseModelValidation, UniqueCodeValidationMixin, ObjectExistsValidationMixin
from invoice.models import PaymentInvoice
from invoice.utils import resolve_payment_details
from invoice.validation.paymentStatusValidation import InvoicePaymentReceiveStatusValidator, \
    InvoicePaymentRefundStatusValidator, InvoicePaymentCancelStatusValidator


class PaymentInvoiceModelValidation(BaseModelValidation, UniqueCodeValidationMixin, ObjectExistsValidationMixin):
    OBJECT_TYPE = PaymentInvoice

    AMOUNT_PAYED_NOT_MATCHING_ITEMS = _(
        "Amount payed for payment_invoice %(payment_invoice)s is not matching amount total. "
        "Expected: %(expected).2f, Payed: %(payed).2f"
    )

    INVALID_AMOUNT_TOTAL = _("Not all payment_invoice line items for payment_invoice %(payment_invoice)s have a amount total attached. "
                             "It's not possible to validate payment. ")

    @classmethod
    def validate_create(cls, user, **data):
        cls.validate_unique_code_name(data.get('code_ext', None))

    @classmethod
    def validate_ref_received(cls, user, payment, payment_ref):
        excluded = payment.id if payment else None
        cls.validate_unique_code_name(payment_ref, excluded_id=excluded, code_key='code_ext')

    @classmethod
    def validate_update(cls, user, **data):
        cls.validate_object_exists(data.get('id', None))
        code = data.get('code_ext', None)
        id_ = data.get('id', None)

        if code:
            cls.validate_unique_code_name(code, id_)

    @classmethod
    def validate_receive_payment(cls, user, invoice_payment):
        cls._validate_payment_match_items(user, invoice_payment)

    @classmethod
    def validate_refund_payment(cls, user, invoice_payment):
        InvoicePaymentRefundStatusValidator(invoice_payment)()

    @classmethod
    def validate_cancel_payment(cls, user, invoice_payment):
        InvoicePaymentCancelStatusValidator(invoice_payment)()

    @classmethod
    def _validate_payment_match_items(cls, user, payment_invoice: PaymentInvoice):
        invoices, bills = resolve_payment_details(payment_invoice)
        line_items = []
        for invoice in invoices:
            line_items += [x.amount_total for x in invoice.line_items.all()]
        for bill in bills:
            line_items += [x.amount_total for x in bill.line_items_bill.all()]
        if len(line_items) == 0:
            raise ValidationError(cls.INVALID_AMOUNT_TOTAL % {
                'payment_invoice': payment_invoice
            })
        if None in line_items:
            raise ValidationError(cls.INVALID_AMOUNT_TOTAL % {
                'payment_invoice': payment_invoice
            })
        expected_amount = sum(line_items)
        if expected_amount != payment_invoice.amount_received:
            raise ValidationError(cls.AMOUNT_PAYED_NOT_MATCHING_ITEMS % {
                'payment_invoice': payment_invoice,
                'expected': expected_amount,
                'payed': payment_invoice.amount_received
            })



# Generated by Django 3.0.14 on 2022-04-13 08:24

import core.models
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0016_add_last_login_on_interactive_user'),
        ('invoice', '0008_auto_20220411_1358'),
    ]

    operations = [
        migrations.AlterField(
            model_name='historicalpaymentinvoice',
            name='reconciliation_status',
            field=models.SmallIntegerField(choices=[(0, 'not reconciliated'), (1, 'reconciliated'), (2, 'refunded'), (3, 'cancelled')], db_column='ReconciliationStatus'),
        ),
        migrations.AlterField(
            model_name='paymentinvoice',
            name='reconciliation_status',
            field=models.SmallIntegerField(choices=[(0, 'not reconciliated'), (1, 'reconciliated'), (2, 'refunded'), (3, 'cancelled')], db_column='ReconciliationStatus'),
        ),
        migrations.CreateModel(
            name='PaymentInvoiceMutation',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('mutation', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='payment_invoices', to='core.MutationLog')),
                ('payment_invoice', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='mutations', to='invoice.PaymentInvoice')),
            ],
            options={
                'db_table': 'paymentinvoice_PaymentInvoiceMutation',
                'managed': True,
            },
            bases=(models.Model, core.models.ObjectMutation),
        ),
        migrations.CreateModel(
            name='DetailPaymentInvoiceMutation',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('detail_payment_invoice', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='mutations', to='invoice.DetailPaymentInvoice')),
                ('mutation', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='detail_payment_invoices', to='core.MutationLog')),
            ],
            options={
                'db_table': 'paymentinvoice_DetailPaymentInvoiceMutation',
                'managed': True,
            },
            bases=(models.Model, core.models.ObjectMutation),
        ),
    ]



# Generated by Django 3.0.14 on 2021-12-03 10:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('invoice', '0002_auto_20211117_0904'),
    ]

    operations = [
        migrations.AddField(
            model_name='billpayment',
            name='payment_origin',
            field=models.CharField(db_column='PaymentOrigin', max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='historicalbillpayment',
            name='payment_origin',
            field=models.CharField(db_column='PaymentOrigin', max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='historicalinvoicepayment',
            name='payment_origin',
            field=models.CharField(db_column='PaymentOrigin', max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='invoicepayment',
            name='payment_origin',
            field=models.CharField(db_column='PaymentOrigin', max_length=255, null=True),
        ),
    ]



from django.db import transaction

from invoice.models import InvoicePayment, Invoice
from core.services import BaseService
from core.services.utils import check_authentication, output_exception
from invoice.validation.invoicePayment import InvoicePaymentModelValidation


class InvoicePaymentsService(BaseService):

    OBJECT_TYPE = InvoicePayment

    def __init__(self, user, validation_class: InvoicePaymentModelValidation = InvoicePaymentModelValidation):
        super().__init__(user, validation_class)
        self.validation_class = validation_class

    @check_authentication
    def update(self, obj_data):
        raise NotImplementedError("Update method is not implemented for InvoicePayment")

    @check_authentication
    def create(self, obj_data):
        raise NotImplementedError("Create method is not implemented for InvoicePayment")

    @check_authentication
    def ref_received(self, invoice_payment: InvoicePayment, payment_ref):
        try:
            with transaction.atomic():
                self.validation_class.validate_ref_received(self.user, invoice_payment, payment_ref)
                invoice_payment.code_ext = payment_ref
                return self.save_instance(invoice_payment)
        except Exception as exc:
            return output_exception(model_name="InvoicePayment", method="ref_received", exception=exc)

    def payment_received(self, invoice_payment: InvoicePayment, payment_status: InvoicePayment.PaymentStatus):
        try:
            with transaction.atomic():
                invoice_payment.status = payment_status
                self.validation_class.validate_receive_payment(self.user, invoice_payment)

                self._update_invoice_status(invoice_payment.invoice, Invoice.Status.PAID)

                invoice_payment.invoice.save(username=self.user.username)
                return self.save_instance(invoice_payment)
        except Exception as exc:
            return output_exception(model_name="InvoicePayment", method="payment_received", exception=exc)

    def payment_refunded(self, invoice_payment):
        try:
            with transaction.atomic():
                self.validation_class.validate_refund_payment(self.user, invoice_payment)
                self._update_payment_status(invoice_payment, InvoicePayment.PaymentStatus.REFUNDED)
                self._update_invoice_status(invoice_payment.invoice, Invoice.Status.SUSPENDED)

                invoice_payment.invoice.save(username=self.user.username)
                return self.save_instance(invoice_payment)
        except Exception as exc:
            return output_exception(model_name="InvoicePayment", method="payment_refunded", exception=exc)

    def payment_cancelled(self, invoice_payment):
        try:
            with transaction.atomic():
                self.validation_class.validate_cancel_payment(self.user, invoice_payment)

                self._update_payment_status(invoice_payment.invoice, InvoicePayment.PaymentStatus.CANCELLED)
                self._update_invoice_status(invoice_payment.invoice, Invoice.Status.SUSPENDED)

                invoice_payment.invoice.save(username=self.user.username)
                return self.save_instance(invoice_payment)
        except Exception as exc:
            return output_exception(model_name="InvoicePayment", method="payment_refunded", exception=exc)

    def _update_payment_status(self, invoice_payment: InvoicePayment, status: InvoicePayment.PaymentStatus):
        invoice_payment.status = status

    def _update_invoice_status(self, invoice_payment: Invoice, status: Invoice.Status):
        invoice_payment.status = status




class GenericFilterGQLTypeMixin:

    @classmethod
    def get_base_filters_invoice(cls):
        return {
            "id": ["exact"],
            "subject_id": ["exact"],
            "subject_type": ["exact"],

            "thirdparty_id": ["exact"],
            "thirdparty_type": ["exact"],
            "code_tp": ["exact", "istartswith", "icontains", "iexact"],
            "code": ["exact", "istartswith", "icontains", "iexact"],
            "code_ext": ["exact", "istartswith", "icontains", "iexact"],
            "date_due": ["exact", "lt", "lte", "gt", "gte"],
            "date_payed": ["exact", "lt", "lte", "gt", "gte"],

            "amount_discount": ["exact", "lt", "lte", "gt", "gte"],
            "amount_net": ["exact", "lt", "lte", "gt", "gte"],
            "amount_total": ["exact", "lt", "lte", "gt", "gte"],

            "status": ["exact"],

            "currency_tp_code": ["exact"],
            "currency_code": ["exact"],
            "note": ["exact", "icontains"],
            "terms": ["exact", "icontains"],

            "payment_reference": ["exact", "istartswith", "icontains", "iexact"],

            "date_created": ["exact", "lt", "lte", "gt", "gte"],
            "date_updated": ["exact", "lt", "lte", "gt", "gte"],
            "is_deleted": ["exact"],
            "version": ["exact"],
        }

    @classmethod
    def get_base_filters_invoice_line_item(cls):
        return {
            "id": ["exact"],
            "code": ["exact", "icontains", "istartswith"],

            "line_type": ["exact"],
            "line_id": ["exact"],

            "description": ["istartswith", "icontains", "iexact"],

            "ledger_account": ["istartswith", "iexact", "icontains"],

            "quantity": ["exact", "lt", "lte", "gt", "gte"],
            "unit_price": ["exact", "lt", "lte", "gt", "gte"],
            "deduction": ["exact", "lt", "lte", "gt", "gte"],
            "discount": ["exact", "lt", "lte", "gt", "gte"],

            "tax_rate": ["exact"],
            "amount_total": ["exact", "lt", "lte", "gt", "gte"],
            "amount_net": ["exact", "lt", "lte", "gt", "gte"],

            "date_created": ["exact", "lt", "lte", "gt", "gte"],
            "date_updated": ["exact", "lt", "lte", "gt", "gte"],
            "is_deleted": ["exact"],
            "version": ["exact"],
        }

    @classmethod
    def get_base_filters_invoice_payment(cls):
        return {
            "id": ["exact"],
            "status": ["exact"],

            "code_tp": ["istartswith", "icontains", "iexact"],
            "code_ext": ["istartswith", "icontains", "iexact"],
            "code_receipt": ["istartswith", "icontains", "iexact"],

            "label": ["istartswith", "iexact"],

            "amount_payed": ["exact", "lt", "lte", "gt", "gte"],
            "fees": ["exact", "lt", "lte", "gt", "gte"],
            "amount_received": ["exact", "lt", "lte", "gt", "gte"],

            "date_payment": ["exact", "lt", "lte", "gt", "gte"],

            "date_created": ["exact", "lt", "lte", "gt", "gte"],
            "date_updated": ["exact", "lt", "lte", "gt", "gte"],

            "payment_origin": ["istartswith", "icontains", "iexact"],

            "is_deleted": ["exact"],
            "version": ["exact"],
        }

    @classmethod
    def get_base_filters_invoice_event(cls):
        return {
            "id": ["exact"],
            "event_type": ["exact", "lt", "lte", "gt", "gte"],
            "message": ["istartswith", "icontains", "iexact"],

            "date_created": ["exact", "lt", "lte", "gt", "gte"],
            "date_updated": ["exact", "lt", "lte", "gt", "gte"],
            "is_deleted": ["exact"],
            "version": ["exact"],
        }

    @classmethod
    def get_base_filters_payment_invoice(cls):
        return {
            "id": ["exact"],
            "code_tp": ["istartswith", "icontains", "iexact"],
            "code_ext": ["istartswith", "icontains", "iexact"],
            "code_receipt": ["istartswith", "icontains", "iexact"],

            "label": ["istartswith", "iexact"],

            "reconciliation_status": ["exact"],

            "fees": ["exact", "lt", "lte", "gt", "gte"],
            "amount_received": ["exact", "lt", "lte", "gt", "gte"],

            "date_payment": ["exact", "lt", "lte", "gt", "gte"],

            "payment_origin": ["istartswith", "icontains", "iexact"],

            "date_created": ["exact", "lt", "lte", "gt", "gte"],
            "date_updated": ["exact", "lt", "lte", "gt", "gte"],

            "payer_ref": ["istartswith", "icontains", "iexact"],
            "payer_name": ["istartswith", "icontains", "iexact"],

            "is_deleted": ["exact"],
            "version": ["exact"],
        }

    @classmethod
    def get_base_filters_detail_invoice_payment(cls):
        return {
            "id": ["exact"],
            "subject_id": ["exact"],
            "subject_type": ["exact"],

            "status": ["exact"],
            "fees": ["exact", "lt", "lte", "gt", "gte"],
            "amount": ["exact", "lt", "lte", "gt", "gte"],

            "reconcilation_id": ["istartswith", "icontains", "iexact"],
            "reconcilation_date": ["exact", "lt", "lte", "gt", "gte"],

            "date_created": ["exact", "lt", "lte", "gt", "gte"],
            "date_updated": ["exact", "lt", "lte", "gt", "gte"],

            "is_deleted": ["exact"],
            "version": ["exact"],
        }



# Generated by Django 3.2.19 on 2023-09-21 08:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('invoice', '0011_auto_20230914_0805'),
    ]

    operations = [
        migrations.AlterField(
            model_name='bill',
            name='status',
            field=models.SmallIntegerField(choices=[(0, 'draft'), (1, 'validated'), (2, 'paid'), (3, 'cancelled'), (4, 'deleted'), (5, 'suspended'), (6, 'unpaid'), (7, 'reconciliated')], db_column='Status', default=0),
        ),
        migrations.AlterField(
            model_name='historicalbill',
            name='status',
            field=models.SmallIntegerField(choices=[(0, 'draft'), (1, 'validated'), (2, 'paid'), (3, 'cancelled'), (4, 'deleted'), (5, 'suspended'), (6, 'unpaid'), (7, 'reconciliated')], db_column='Status', default=0),
        ),
        migrations.AlterField(
            model_name='historicalinvoice',
            name='status',
            field=models.SmallIntegerField(choices=[(0, 'draft'), (1, 'validated'), (2, 'paid'), (3, 'cancelled'), (4, 'deleted'), (5, 'suspended'), (6, 'unpaid'), (7, 'reconciliated')], db_column='Status', default=0),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='status',
            field=models.SmallIntegerField(choices=[(0, 'draft'), (1, 'validated'), (2, 'paid'), (3, 'cancelled'), (4, 'deleted'), (5, 'suspended'), (6, 'unpaid'), (7, 'reconciliated')], db_column='Status', default=0),
        ),
    ]



import logging

from django.db import migrations

from core.utils import insert_role_right_for_system

logger = logging.getLogger(__name__)


def add_rights(apps, schema_editor):
    insert_role_right_for_system(64, 155101)  # Invoice search
    insert_role_right_for_system(64, 155102)  # Invoice create
    insert_role_right_for_system(64, 155103)  # Invoice update
    insert_role_right_for_system(64, 155104)  # Invoice delete
    insert_role_right_for_system(64, 155109)  # Invoice amend
    insert_role_right_for_system(64, 155201)  # Invoice payment search
    insert_role_right_for_system(64, 155202)  # Invoice payment create
    insert_role_right_for_system(64, 155203)  # Invoice payment update
    insert_role_right_for_system(64, 155204)  # Invoice payment delete
    insert_role_right_for_system(64, 155206)  # Invoice payment refund
    insert_role_right_for_system(64, 155301)  # Invoice Event search
    insert_role_right_for_system(64, 155302)  # Invoice Event create
    insert_role_right_for_system(64, 155303)  # Invoice Event update
    insert_role_right_for_system(64, 155304)  # Invoice Event delete
    insert_role_right_for_system(64, 155306)  # Invoice Event message
    insert_role_right_for_system(64, 155307)  # Invoice Event delete my message
    insert_role_right_for_system(64, 155308)  # Invoice Event delete all messages
    insert_role_right_for_system(64, 156101)  # Bill search
    insert_role_right_for_system(64, 156102)  # Bill create
    insert_role_right_for_system(64, 156103)  # Bill update
    insert_role_right_for_system(64, 156104)  # Bill delete
    insert_role_right_for_system(64, 156109)  # Bill amend
    insert_role_right_for_system(64, 156201)  # Bill Payment search
    insert_role_right_for_system(64, 156202)  # Bill Payment create
    insert_role_right_for_system(64, 156203)  # Bill Payment update
    insert_role_right_for_system(64, 156204)  # Bill Payment delete
    insert_role_right_for_system(64, 156206)  # Bill Payment refund
    insert_role_right_for_system(64, 156301)  # Bill Event search
    insert_role_right_for_system(64, 156302)  # Bill Event create
    insert_role_right_for_system(64, 156303)  # Bill Event update
    insert_role_right_for_system(64, 156304)  # Bill Event delete
    insert_role_right_for_system(64, 156306)  # Bill Event create message
    insert_role_right_for_system(64, 156307)  # Bill Event delete my message
    insert_role_right_for_system(64, 156308)  # Bill Event delete all messages


class Migration(migrations.Migration):
    dependencies = [
        ('invoice', '0003_auto_20211203_1053')
    ]

    operations = [
        migrations.RunPython(add_rights),
    ]



from datetime import date

from django.contrib.contenttypes.models import ContentType

from contract.tests.helpers import create_test_contract
from policyholder.tests.helpers import create_test_policy_holder
from insuree.test_helpers import create_test_insuree
from core.forms import User

from invoice.models import Invoice

DEFAULT_TEST_INVOICE_PAYLOAD = {
    'subject_type': 'contract',
    'subject_id': None,
    'thirdparty_type': 'insuree',
    'thirdparty_id': None,
    'code': 'INVOICE_CODE',
    'code_tp': 'INVOICE_CODE_TP',
    'code_ext': 'INVOICE_CODE_EXT',
    'date_due': date(2021, 9, 13),
    'date_invoice': date(2021, 9, 11),
    'date_payed': date(2021, 9, 12),
    'amount_discount': 20.1,
    'amount_net': 20.1,
    'tax_analysis': {'lines': [{'code': 'c', 'label': 'l', 'base': '0.1', 'amount': '2.01'}], 'total': '2.01'},
    'amount_total': 20.1,
    'status': 0,  # Draft
    'note': 'NOTE',
    'terms': 'TERMS',
    'payment_reference': 'payment reference'
}


def __get_or_create_user():
    user = User.objects.filter(username='admin_invoice').first()
    if not user:
        user = User.objects.create_superuser(username='admin_invoice', password='S\/pe®Pąßw0rd™')
    return user


def create_test_invoice(subject=None, thirdparty=None, user=None, **custom_props):
    subject = subject or __create_test_subject()
    thirdparty = thirdparty or __create_test_thirdparty()
    payload = DEFAULT_TEST_INVOICE_PAYLOAD.copy()
    payload['subject'] = subject
    payload['subject_type'] = ContentType.objects.get_for_model(subject)
    payload['thirdparty'] = thirdparty
    payload['thirdparty_type'] = ContentType.objects.get_for_model(thirdparty)
    payload.update(**custom_props)

    Invoice.objects.filter(code=payload['code']).delete()

    user = user or __get_or_create_user()
    invoice = Invoice(**payload)
    invoice.save(username=user.username)

    return invoice


def __create_test_subject():
    policy_holder = create_test_policy_holder()
    return create_test_contract(policy_holder)


def __create_test_thirdparty():
    return create_test_insuree(with_family=False)
